# Copyright (c) 2018 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

# pylint: disable=wrong-import-position

import os as _os
import sys as _sys

__version__ = "1.15.1"
__version_info__ = tuple(map(int, __version__.split(".")[:3]))
__license__ = "MIT"
__author__ = "UAVCAN Consortium"
__copyright__ = "Copyright (c) 2018 UAVCAN Consortium"
__email__ = "consortium@uavcan.org"

# Our unorthodox approach to dependency management requires us to apply certain workarounds.
# Here, the objective is to allow our library to import stuff from its third-party dependency directory,
# but at the same time we don't want to interfere with the application that depends on this library.
# So we modify the import lookup path temporarily while the package initialization is in progress;
# when done, we restore the path back to its original value. One implication is that it won't be possible
# to import stuff dynamically after the initialization is finished (e.g., function-local imports won't be
# able to reach the third-party stuff), but we don't care.
_original_sys_path = _sys.path
_sys.path = [_os.path.join(_os.path.dirname(__file__), "third_party")] + _sys.path

# Never import anything that is not available here - API stability guarantees are only provided for the exposed items.
from ._namespace import read_namespace as read_namespace
from ._namespace import PrintOutputHandler as PrintOutputHandler

# Error model.
from ._error import FrontendError as FrontendError
from ._error import InvalidDefinitionError as InvalidDefinitionError
from ._error import InternalError as InternalError

# Data type model - meta types.
from ._serializable import SerializableType as SerializableType
from ._serializable import PrimitiveType as PrimitiveType
from ._serializable import BooleanType as BooleanType
from ._serializable import ArithmeticType as ArithmeticType
from ._serializable import IntegerType as IntegerType
from ._serializable import SignedIntegerType as SignedIntegerType
from ._serializable import UnsignedIntegerType as UnsignedIntegerType
from ._serializable import FloatType as FloatType
from ._serializable import VoidType as VoidType
from ._serializable import ArrayType as ArrayType
from ._serializable import FixedLengthArrayType as FixedLengthArrayType
from ._serializable import VariableLengthArrayType as VariableLengthArrayType
from ._serializable import CompositeType as CompositeType
from ._serializable import UnionType as UnionType
from ._serializable import StructureType as StructureType
from ._serializable import DelimitedType as DelimitedType
from ._serializable import ServiceType as ServiceType

# Data type model - attributes.
from ._serializable import Attribute as Attribute
from ._serializable import Field as Field
from ._serializable import PaddingField as PaddingField
from ._serializable import Constant as Constant

# Expression model.
from ._expression import Any as Any
from ._expression import Primitive as Primitive
from ._expression import Boolean as Boolean
from ._expression import Rational as Rational
from ._expression import String as String
from ._expression import Container as Container
from ._expression import Set as Set

# Auxiliary.
from ._serializable import ValueRange as ValueRange
from ._serializable import Version as Version
from ._bit_length_set import BitLengthSet as BitLengthSet

_sys.path = _original_sys_path



# Copyright (c) 2018 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from .. import _expression
from ._serializable import SerializableType, TypeParameterError
from ._primitive import UnsignedIntegerType, PrimitiveType, FloatType, ArithmeticType, IntegerType, BooleanType
from ._void import VoidType
from ._name import check_name, InvalidNameError


class InvalidConstantValueError(TypeParameterError):
    pass


class InvalidTypeError(TypeParameterError):
    pass


class Attribute(_expression.Any):
    def __init__(self, data_type: SerializableType, name: str, doc: str = ""):
        self._data_type = data_type
        self._name = str(name)
        self._doc = str(doc)

        if isinstance(data_type, VoidType):
            if self._name:
                raise InvalidNameError("Void-typed fields can be used only for padding and cannot be named")
        else:
            check_name(self._name)

    @property
    def data_type(self) -> SerializableType:
        return self._data_type

    @property
    def name(self) -> str:
        """For padding fields this is an empty string."""
        return self._name

    @property
    def doc(self) -> str:
        """Docs for this attribute without the leading #."""
        return self._doc

    def __hash__(self) -> int:
        return hash((self._data_type, self._name))

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Attribute):
            return (self._data_type == other._data_type) and (self._name == other.name)
        return NotImplemented  # pragma: no cover

    def __str__(self) -> str:
        """Returns the normalized DSDL representation of the attribute."""
        return ("%s %s" % (self.data_type, self.name)).strip()

    def __repr__(self) -> str:
        return "%s(data_type=%r, name=%r)" % (self.__class__.__name__, self.data_type, self.name)


class Field(Attribute):
    pass


class PaddingField(Field):
    def __init__(self, data_type: VoidType, doc: str = ""):
        if not isinstance(data_type, VoidType):
            raise TypeParameterError("Padding fields must be of the void type")

        super().__init__(data_type, "", doc)


class Constant(Attribute):
    def __init__(self, data_type: SerializableType, name: str, value: _expression.Any, doc: str = ""):
        super().__init__(data_type, name, doc)

        if not isinstance(value, _expression.Primitive):
            raise InvalidConstantValueError("The constant value must be a primitive expression value")

        self._value = value
        del value

        # Interestingly, both the type of the constant and its value are instances of the same meta-type: expression.
        # BooleanType inherits from expression.Any, same as expression.Boolean.
        if isinstance(data_type, BooleanType):  # Boolean constant
            if not isinstance(self._value, _expression.Boolean):
                raise InvalidConstantValueError("Invalid value for boolean constant: %r" % self._value)

        elif isinstance(data_type, IntegerType):  # Integer constant
            if isinstance(self._value, _expression.Rational):
                if not self._value.is_integer():
                    raise InvalidConstantValueError(
                        "The value of an integer constant must be an integer; got %s" % self._value
                    )
            elif isinstance(self._value, _expression.String):
                as_bytes = self._value.native_value.encode("utf8")
                if len(as_bytes) != 1:
                    raise InvalidConstantValueError("A constant string must be exactly one ASCII character long")

                if not isinstance(data_type, UnsignedIntegerType) or data_type.bit_length != 8:
                    raise InvalidConstantValueError("Constant strings can be used only with uint8")

                self._value = _expression.Rational(ord(as_bytes))  # Replace string with integer
            else:
                raise InvalidConstantValueError("Invalid value type for integer constant: %r" % self._value)

        elif isinstance(data_type, FloatType):  # Floating point constant
            if not isinstance(self._value, _expression.Rational):
                raise InvalidConstantValueError("Invalid value type for float constant: %r" % self._value)

        else:
            raise InvalidTypeError("Invalid constant type: %r" % data_type)

        assert isinstance(self._value, _expression.Any)
        assert isinstance(self._value, _expression.Rational) == isinstance(self.data_type, (FloatType, IntegerType))
        assert isinstance(self._value, _expression.Boolean) == isinstance(self.data_type, BooleanType)

        # Range check
        if isinstance(self._value, _expression.Rational):
            assert isinstance(data_type, ArithmeticType)
            rng = data_type.inclusive_value_range
            if not (rng.min <= self._value.native_value <= rng.max):
                raise InvalidConstantValueError(
                    "Constant value %s exceeds the range of its data type %s" % (self._value, data_type)
                )

    @property
    def value(self) -> _expression.Any:
        """
        The result of evaluating the constant initialization expression.
        The value is guaranteed to be compliant with the constant's own type -- it is checked at the evaluation time.
        The compliance rules are defined in the Specification.
        """
        return self._value

    def __hash__(self) -> int:
        return hash((self._data_type, self._name, self._value))

    def __eq__(self, other: object) -> bool:
        """Constants are equal if their type, name, and value are equal."""
        if isinstance(other, Constant):
            return super().__eq__(other) and (self._value == other._value)
        return NotImplemented  # pragma: no cover

    def __str__(self) -> str:
        """Returns the normalized DSDL representation of the constant and its value."""
        return "%s %s = %s" % (self.data_type, self.name, self.value)

    def __repr__(self) -> str:
        return "Constant(data_type=%r, name=%r, value=%r)" % (self.data_type, self.name, self._value)


def _unittest_attribute() -> None:
    from pytest import raises
    from ._primitive import SignedIntegerType

    assert str(Field(BooleanType(PrimitiveType.CastMode.SATURATED), "flag")) == "saturated bool flag"
    assert (
        repr(Field(BooleanType(PrimitiveType.CastMode.SATURATED), "flag"))
        == "Field(data_type=BooleanType(bit_length=1, cast_mode=<CastMode.SATURATED: 0>), name='flag')"
    )

    assert str(PaddingField(VoidType(32))) == "void32"
    assert repr(PaddingField(VoidType(1))) == "PaddingField(data_type=VoidType(bit_length=1), name='')"

    assert Field(UnsignedIntegerType(1, PrimitiveType.CastMode.SATURATED), "flag") == Field(
        UnsignedIntegerType(1, PrimitiveType.CastMode.SATURATED), "flag"
    )
    assert hash(Field(UnsignedIntegerType(1, PrimitiveType.CastMode.SATURATED), "flag")) == hash(
        Field(UnsignedIntegerType(1, PrimitiveType.CastMode.SATURATED), "flag")
    )

    assert Field(UnsignedIntegerType(1, PrimitiveType.CastMode.TRUNCATED), "flag") != Field(
        UnsignedIntegerType(1, PrimitiveType.CastMode.SATURATED), "flag"
    )
    assert hash(Field(UnsignedIntegerType(1, PrimitiveType.CastMode.TRUNCATED), "flag")) != hash(
        Field(UnsignedIntegerType(1, PrimitiveType.CastMode.SATURATED), "flag")
    )

    with raises(TypeParameterError, match=".*void.*"):
        # noinspection PyTypeChecker
        repr(PaddingField(SignedIntegerType(8, PrimitiveType.CastMode.SATURATED)))  # type: ignore

    data_type = SignedIntegerType(32, PrimitiveType.CastMode.SATURATED)
    const = Constant(data_type, "FOO_CONST", _expression.Rational(-123))
    assert str(const) == "saturated int32 FOO_CONST = -123"
    assert const.data_type is data_type
    assert const.name == "FOO_CONST"
    assert const.value == _expression.Rational(-123)

    assert repr(const) == "Constant(data_type=%r, name='FOO_CONST', value=rational(-123))" % data_type

    assert Constant(data_type, "FOO_CONST", _expression.Rational(-123)) == const
    assert Constant(data_type, "FOO_CONST", _expression.Rational(-124)) != const
    assert hash(Constant(data_type, "FOO_CONST", _expression.Rational(-123))) == hash(const)
    assert hash(Constant(data_type, "FOO_CONST", _expression.Rational(-124))) != hash(const)



# Copyright (c) 2018 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

# pylint: disable=global-statement,protected-access,too-many-statements,consider-using-with

import os
import typing
import tempfile
from typing import Union, Tuple, Optional
from pathlib import Path
from textwrap import dedent
from . import _expression
from . import _error
from . import _parser
from . import _data_type_builder
from . import _dsdl_definition
from . import _serializable
from . import _namespace


# Type annotation disabled here because MyPy is misbehaving, reporting these nonsensical error messages:
#   pydsdl/_test.py:18: error: Missing type parameters for generic type
#   pydsdl/_test.py: note: In function "_in_n_out":
#   pydsdl/_test.py:18: error: Missing type parameters for generic type
_DIRECTORY = None  # type : typing.Optional[tempfile.TemporaryDirectory]


def _parse_definition(
    definition: _dsdl_definition.DSDLDefinition, lookup_definitions: typing.Sequence[_dsdl_definition.DSDLDefinition]
) -> _serializable.CompositeType:
    return definition.read(
        lookup_definitions,
        print_output_handler=lambda line, text: print("Output from line %d:" % line, text),
        allow_unregulated_fixed_port_id=False,
    )


def _define(rel_path: Union[str, Path], text: str) -> _dsdl_definition.DSDLDefinition:
    rel_path = str(rel_path).replace("/", os.sep)  # Windows compatibility
    assert _DIRECTORY
    path = Path(_DIRECTORY.name, rel_path)
    os.makedirs(os.path.dirname(path), exist_ok=True)
    with open(path, "w", encoding="utf8") as f:
        f.write(text)

    root_namespace_path = Path(_DIRECTORY.name, rel_path.strip(os.sep).split(os.sep, maxsplit=1)[0])
    out = _dsdl_definition.DSDLDefinition(path, root_namespace_path)
    print("New definition:", out, "Root NS:", root_namespace_path)
    return out


def _in_n_out(test: typing.Callable[[], None]) -> typing.Callable[[], None]:
    def decorator() -> None:
        global _DIRECTORY
        _DIRECTORY = tempfile.TemporaryDirectory(prefix="pydsdl-test-")
        try:
            test()
        finally:
            _DIRECTORY = None  # Preserving the contents for future inspection if needed

    return decorator


@_in_n_out
def _unittest_define() -> None:
    # I DON'T ALWAYS WRITE UNIT TESTS
    d = _define("uavcan/test/5000.Message.1.2.uavcan", "# empty")
    assert _DIRECTORY is not None
    assert d.full_name == "uavcan.test.Message"
    assert d.version == (1, 2)
    assert d.fixed_port_id == 5000
    assert d.file_path == Path(_DIRECTORY.name, "uavcan", "test", "5000.Message.1.2.uavcan").resolve()
    assert d.root_namespace_path == Path(_DIRECTORY.name, "uavcan").resolve()
    assert open(d.file_path).read() == "# empty"

    # BUT WHEN I DO, I WRITE UNIT TESTS FOR MY UNIT TESTS
    d = _define("uavcan/Service.255.254.uavcan", "# empty 2")
    assert d.full_name == "uavcan.Service"
    assert d.version == (255, 254)
    assert d.fixed_port_id is None
    assert d.file_path == Path(_DIRECTORY.name, "uavcan", "Service.255.254.uavcan").resolve()
    assert d.root_namespace_path == Path(_DIRECTORY.name, "uavcan").resolve()
    assert open(d.file_path).read() == "# empty 2"


@_in_n_out
def _unittest_simple() -> None:
    abc = _define(
        "vendor/nested/7000.Abc.1.2.uavcan",
        dedent(
            """
        @deprecated
        uint8 CHARACTER = '#'
        int8 a
        saturated int64[<33] b
        @extent 1024 * 8
        """
        ),
    )
    assert abc.fixed_port_id == 7000
    assert abc.full_name == "vendor.nested.Abc"
    assert abc.version == (1, 2)

    p = _parse_definition(abc, [])
    print("Parsed:", p)
    assert isinstance(p, _serializable.DelimitedType)
    assert isinstance(p.inner_type, _serializable.StructureType)
    assert p.full_name == "vendor.nested.Abc"
    assert str(p.source_file_path).endswith(os.path.join("vendor", "nested", "7000.Abc.1.2.uavcan"))
    assert p.source_file_path == abc.file_path
    assert p.fixed_port_id == 7000
    assert p.deprecated
    assert p.version == (1, 2)
    assert min(p.inner_type.bit_length_set) == 16
    assert max(p.inner_type.bit_length_set) == 16 + 64 * 32
    assert min(p.bit_length_set) == 32
    assert max(p.bit_length_set) == 32 + 1024 * 8
    assert len(p.attributes) == 3
    assert len(p.fields) == 2
    assert str(p.fields[0].data_type) == "saturated int8"
    assert p.fields[0].name == "a"
    assert str(p.fields[1].data_type) == "saturated int64[<=32]"  # Note: normalized representation
    assert p.fields[1].name == "b"
    assert len(p.constants) == 1
    assert str(p.constants[0].data_type) == "saturated uint8"
    assert p.constants[0].name == "CHARACTER"
    assert isinstance(p.constants[0].value, _expression.Rational)
    assert p.constants[0].value == _expression.Rational(ord("#"))

    t = p.fields[1].data_type
    assert isinstance(t, _serializable.ArrayType)
    assert str(t.element_type) == "saturated int64"

    empty_new = _define("vendor/nested/Empty.255.255.uavcan", """@sealed""")

    empty_old = _define("vendor/nested/Empty.255.254.uavcan", """@sealed""")

    constants = _define(
        "another/Constants.5.0.uavcan",
        dedent(
            """
        @sealed
        float64 PI = 3.1415926535897932384626433
        """
        ),
    )

    service = _define(
        "another/300.Service.0.1.uavcan",
        dedent(
            """
        @union
        @deprecated
        vendor.nested.Empty.255.255 new_empty_implicit
        vendor.nested.Empty.255.255 new_empty_explicit
        vendor.nested.Empty.255.254 old_empty
        @extent 32
        -----------------------------------
        @sealed                      # RESPONSE SEALED REQUEST NOT
        Constants.5.0 constants      # RELATIVE REFERENCE
        vendor.nested.Abc.1.2 abc
        """
        ),
    )

    p = _parse_definition(
        service,
        [
            abc,
            empty_new,
            empty_old,
            constants,
        ],
    )
    print("Parsed:", p)
    assert isinstance(p, _serializable.ServiceType)
    assert p.full_name == "another.Service"
    assert p.fixed_port_id == 300
    assert p.deprecated
    assert p.version == (0, 1)
    assert not p.constants
    assert p == p  # pylint: disable=comparison-with-itself
    assert p != empty_new

    assert len(p.fields) == 2
    assert p.fields[0].name == "request"
    assert p.fields[1].name == "response"
    req, res = [x.data_type for x in p.fields]
    assert isinstance(req, _serializable.DelimitedType)
    assert isinstance(req.inner_type, _serializable.UnionType)
    assert isinstance(res, _serializable.StructureType)
    assert req.full_name == "another.Service.Request"
    assert res.full_name == "another.Service.Response"
    assert req is p.request_type
    assert res is p.response_type
    assert req.has_parent_service
    assert req.inner_type.has_parent_service
    assert res.has_parent_service

    assert len(req.constants) == 0
    assert len(req.fields) == 3
    # noinspection PyUnresolvedReferences
    assert req.inner_type.number_of_variants == 3
    assert req.deprecated
    assert not req.has_fixed_port_id
    assert req.version == (0, 1)
    assert req.bit_length_set == {32, 32 + 8, 32 + 16, 32 + 24, 32 + 32}  # Delimited container, @extent
    assert req.inner_type.bit_length_set == 8  # Remember this is a union
    assert [x.name for x in req.fields] == ["new_empty_implicit", "new_empty_explicit", "old_empty"]

    t = req.fields[0].data_type
    assert isinstance(t, _serializable.StructureType)
    assert t.full_name == "vendor.nested.Empty"
    assert t.version == (255, 255)  # Selected implicitly

    t = req.fields[1].data_type
    assert isinstance(t, _serializable.StructureType)
    assert t.full_name == "vendor.nested.Empty"
    assert t.version == (255, 255)  # Selected explicitly

    t = req.fields[2].data_type
    assert isinstance(t, _serializable.StructureType)
    assert t.full_name == "vendor.nested.Empty"
    assert t.version == (255, 254)  # Selected explicitly

    assert len(res.constants) == 0
    assert len(res.fields) == 2
    assert res.deprecated
    assert not res.has_fixed_port_id
    assert res.version == (0, 1)
    # This is a sealed type, so we get the real BLS, but we mustn't forget about the non-sealed nested field!
    assert min(res.bit_length_set) == 32  # Just the delimiter header
    assert max(res.bit_length_set) == 32 + 1024 * 8

    t = res.fields[0].data_type
    assert isinstance(t, _serializable.StructureType)
    assert t.full_name == "another.Constants"
    assert t.version == (5, 0)

    t = res.fields[1].data_type
    assert isinstance(t, _serializable.DelimitedType)
    assert isinstance(t.inner_type, _serializable.StructureType)
    assert t.full_name == "vendor.nested.Abc"
    assert t.version == (1, 2)

    p2 = _parse_definition(
        abc,
        [
            service,
            empty_new,
            empty_old,
            constants,
        ],
    )
    assert hash(p2) == hash(p2)
    assert hash(p2) != hash(p)
    assert hash(p) == hash(p)

    union = _define(
        "another/Union.5.9.uavcan",
        dedent(
            """
        @union
        @sealed
        truncated float16 PI = 3.1415926535897932384626433
        uint8 a
        vendor.nested.Empty.255.255[5] b
        saturated bool [ <= 255 ] c
        """
        ),
    )

    p = _parse_definition(
        union,
        [
            empty_old,
            empty_new,
        ],
    )

    assert p.full_name == "another.Union"
    assert p.version == (5, 9)
    assert p.fixed_port_id is None
    assert not p.has_fixed_port_id
    assert not p.deprecated
    assert isinstance(p, _serializable.UnionType)
    assert p.number_of_variants == 3
    assert len(p.constants) == 1
    assert p.constants[0].name == "PI"
    assert str(p.constants[0].data_type) == "truncated float16"
    assert min(p.bit_length_set) == 8
    assert max(p.bit_length_set) == 8 + 8 + 255 + 1  # The last +1 is the padding to byte.
    assert len(p.fields) == 3
    assert str(p.fields[0]) == "saturated uint8 a"
    assert str(p.fields[1]) == "vendor.nested.Empty.255.255[5] b"
    assert str(p.fields[2]) == "saturated bool[<=255] c"


@_in_n_out
def _unittest_comments() -> None:
    abc = _define(
        "vendor/nested/7000.Abc.1.2.uavcan",
        dedent(
            """\
        # header comment here
        # multiline

        # this should be ignored

        uint8 CHARACTER = '#' # comment on constant
        int8 a # comment on field
        int8 aprime
        @assert 1 == 1 # toss one in for confusion
        void2 # comment on padding field
        saturated int64[<33] b
        # comment on array
        # and another
        @extent 1024 * 8
        """
        ),
    )

    p = _parse_definition(abc, [])
    print("Parsed:", p)
    print(p.doc.__repr__())
    # assert p.doc == "header comment here\nmultiline"
    assert p.constants[0].doc == "comment on constant"
    assert p.fields[0].doc == "comment on field"
    assert p.fields[2].doc == "comment on padding field"
    assert p.fields[3].doc == "comment on array\nand another"

    empty_new = _define("vendor/nested/Empty.255.255.uavcan", """@sealed""")

    empty_old = _define("vendor/nested/Empty.255.254.uavcan", """@sealed""")

    constants = _define(
        "another/Constants.5.0.uavcan",
        dedent(
            """
        @sealed
        float64 PI = 3.1415926535897932384626433 # no header comment
        """
        ),
    )

    p = _parse_definition(constants, [])
    assert p.doc == ""
    assert p.constants[0].doc == "no header comment"

    service = _define(
        "another/300.Service.0.1.uavcan",
        dedent(
            """\
        # first header comment here
        # multiline
        @union
        @deprecated
        vendor.nested.Empty.255.255 new_empty_implicit
        vendor.nested.Empty.255.255 new_empty_explicit
        vendor.nested.Empty.255.254 old_empty
        @extent 32 # make sure no leaks
        -----------------------------------
        # second header comment here
        # multiline
        @sealed                      # RESPONSE SEALED REQUEST NOT
        Constants.5.0 constants      # RELATIVE REFERENCE
        vendor.nested.Abc.1.2 abc
        """
        ),
    )

    p = _parse_definition(
        service,
        [
            abc,
            empty_new,
            empty_old,
            constants,
        ],
    )
    print("Parsed:", p)
    req, res = [x.data_type for x in p.fields]
    assert req.doc == "first header comment here\nmultiline"  # type: ignore
    assert res.doc == "second header comment here\nmultiline"  # type: ignore

    union = _define(
        "another/Union.5.9.uavcan",
        dedent(
            """
        @union
        # sandwiched comment has no effect
        @sealed
        truncated float16 PI = 3.1415926535897932384626433
        uint8 a
        vendor.nested.Empty.255.255[5] b
        saturated bool [ <= 255 ] c
        """
        ),
    )

    p = _parse_definition(
        union,
        [
            empty_old,
            empty_new,
        ],
    )

    assert p.constants[0].doc == ""


# noinspection PyProtectedMember,PyProtectedMember
@_in_n_out
def _unittest_error() -> None:
    from pytest import raises

    def standalone(rel_path: str, definition: str, allow_unregulated: bool = False) -> _serializable.CompositeType:
        return _define(rel_path, definition + "\n").read([], lambda *_: None, allow_unregulated)  # pragma: no branch

    with raises(_error.InvalidDefinitionError, match="(?i).*port ID.*"):
        standalone("vendor/1000.InvalidRegulatedSubjectID.1.0.uavcan", "uint2 value\n@sealed")

    with raises(_error.InvalidDefinitionError, match="(?i).*port ID.*"):
        standalone("vendor/10.InvalidRegulatedServiceID.1.0.uavcan", "uint2 v1\n@sealed\n---\nint64 v2\n@sealed")

    with raises(_error.InvalidDefinitionError, match="(?i).*subject ID.*"):
        standalone("vendor/100000.InvalidRegulatedSubjectID.1.0.uavcan", "uint2 value\n@sealed")

    with raises(_error.InvalidDefinitionError, match="(?i).*service ID.*"):
        standalone("vendor/1000.InvalidRegulatedServiceID.1.0.uavcan", "uint2 v1\n@sealed\n---\nint64 v2\n@sealed")

    with raises(_error.InvalidDefinitionError, match="(?i).*multiple attributes under the same name.*"):
        standalone("vendor/AttributeNameCollision.1.0.uavcan", "uint2 value\n@sealed\nint64 value")

    with raises(_error.InvalidDefinitionError, match="(?i).*tagged union cannot contain fewer than.*"):
        standalone("vendor/SmallUnion.1.0.uavcan", "@union\nuint2 value\n@sealed")

    assert (
        standalone("vendor/invalid_constant_value/A.1.0.uavcan", "bool BOOLEAN = false\n@sealed").constants[0].name
        == "BOOLEAN"
    )
    with raises(_error.InvalidDefinitionError, match=".*Invalid value for boolean constant.*"):
        standalone("vendor/invalid_constant_value/A.1.0.uavcan", "bool BOOLEAN = 0\n@extent 0")  # Should be false

    with raises(_error.InvalidDefinitionError, match=".*undefined_identifier.*"):
        standalone("vendor/invalid_constant_value/A.1.0.uavcan", "bool BOOLEAN = undefined_identifier\n@extent 0")

    with raises(_parser.DSDLSyntaxError):
        standalone("vendor/invalid_constant_value/A.1.0.uavcan", "bool BOOLEAN = -\n@extent 0")

    with raises(_error.InvalidDefinitionError, match="(?i).*exceeds the range.*"):
        standalone("vendor/invalid_constant_value/A.1.0.uavcan", "uint10 INTEGRAL = 2000\n@extent 0")

    with raises(_error.InvalidDefinitionError, match="(?i).*character.*"):
        standalone("vendor/invalid_constant_value/A.1.0.uavcan", "uint8 CH = '\u0451'\n@extent 0")

    with raises(_error.InvalidDefinitionError, match=".*uint8.*"):
        standalone("vendor/invalid_constant_value/A.1.0.uavcan", "uint9 CH = 'q'\n@extent 0")

    with raises(_error.InvalidDefinitionError, match=".*uint8.*"):
        standalone("vendor/invalid_constant_value/A.1.0.uavcan", "int8 CH = 'q'\n@extent 0")

    with raises(_error.InvalidDefinitionError, match=".*integer constant.*"):
        standalone("vendor/invalid_constant_value/A.1.0.uavcan", "int8 CH = 1.1\n@extent 0")

    with raises(_error.InvalidDefinitionError, match="(?i).*type.*"):
        standalone("vendor/invalid_constant_value/A.1.0.uavcan", "float32 CH = true\n@extent 0")

    with raises(_error.InvalidDefinitionError, match="(?i).*type.*"):
        standalone("vendor/invalid_constant_value/A.1.0.uavcan", "float32 CH = 't'\n@extent 0")

    with raises(_parser.DSDLSyntaxError):
        standalone("vendor/syntax_error/A.1.0.uavcan", "bool array[10]\n@sealed")

    with raises(_error.InvalidDefinitionError, match="(?i).*array capacity.*"):
        standalone("vendor/array_size/A.1.0.uavcan", "bool[0] array\n@sealed")

    with raises(_error.InvalidDefinitionError, match="(?i).*array capacity.*"):
        standalone("vendor/array_size/A.1.0.uavcan", "bool[<1] array\n@sealed")

    with raises(_error.InvalidDefinitionError, match="(?i).*array capacity.*"):
        standalone("vendor/array_size/A.1.0.uavcan", "bool[true] array\n@sealed")

    with raises(_error.InvalidDefinitionError, match="(?i).*array capacity.*"):
        standalone("vendor/array_size/A.1.0.uavcan", 'bool["text"] array\n@sealed')

    with raises(_error.InvalidDefinitionError, match="(?i).*service response marker.*"):
        standalone(
            "vendor/service/A.1.0.uavcan",
            "bool request\n@sealed\n---\nbool response\n@sealed\n---\nbool again\n@sealed",
        )

    with raises(_error.InvalidDefinitionError, match="(?i).*unknown directive.*"):
        standalone("vendor/directive/A.1.0.uavcan", "@sho_tse_take\n@sealed")

    with raises(_error.InvalidDefinitionError, match="(?i).*requires an expression.*"):
        standalone("vendor/directive/A.1.0.uavcan", "@assert\n@sealed")

    with raises(_error.InvalidDefinitionError, match="(?i).*does not expect an expression.*"):
        standalone("vendor/directive/A.1.0.uavcan", "@union true || false\n@sealed")

    with raises(_error.InvalidDefinitionError, match="(?i).*does not expect an expression.*"):
        standalone("vendor/directive/A.1.0.uavcan", "@deprecated true || false\n@sealed")

    with raises(_error.InvalidDefinitionError, match="(?i).*version number.*"):
        standalone("vendor/version/A.0.0.uavcan", "@sealed")

    with raises(_error.InvalidDefinitionError, match="(?i).*version number.*"):
        standalone("vendor/version/A.0.256.uavcan", "@sealed")

    with raises(_dsdl_definition.FileNameFormatError):
        standalone("vendor/version/A.0..256.uavcan", "@sealed")

    with raises(_error.InvalidDefinitionError, match="(?i).*version number.*"):
        standalone("vendor/version/A.256.0.uavcan", "@sealed")

    with raises(_parser.DSDLSyntaxError):
        standalone("vendor/types/A.1.0.uavcan", "truncated uavcan.node.Heartbeat.1.0 field\n@sealed")

    with raises(_serializable._primitive.InvalidCastModeError):
        standalone("vendor/types/A.1.0.uavcan", "truncated bool foo\n@sealed")

    with raises(_serializable._primitive.InvalidCastModeError):
        standalone("vendor/types/A.1.0.uavcan", "truncated int8 foo\n@sealed")

    with raises(_data_type_builder.UndefinedDataTypeError, match=r"(?i).*nonexistent.TypeName.*1\.0.*"):
        standalone("vendor/types/A.1.0.uavcan", "nonexistent.TypeName.1.0 field\n@sealed")

    with raises(_data_type_builder.UndefinedDataTypeError, match=r"(?i).*vendor[/\\]+types instead of .*vendor.*"):
        standalone("vendor/types/A.1.0.uavcan", "types.Nonexistent.1.0 field\n@sealed")

    with raises(_error.InvalidDefinitionError, match=r"(?i).*not defined for.*"):
        standalone(
            "vendor/types/A.1.0.uavcan",
            dedent(
                """
                   @union
                   int8 a
                   @assert _offset_.count >= 1
                   int16 b
                   @sealed
                   """
            ),
        )

    with raises(_error.InvalidDefinitionError, match=r"(?i).*field offset is not defined for unions.*"):
        standalone(
            "vendor/types/A.1.0.uavcan",
            dedent(
                """
                   @union
                   int8 a
                   int16 b
                   @assert _offset_.count >= 1
                   int8 c
                   @sealed
                   """
            ),
        )

    with raises(_data_type_builder.UndefinedDataTypeError, match=r".*ns.Type_.*1\.0"):
        _parse_definition(
            _define("vendor/types/A.1.0.uavcan", "ns.Type_.1.0 field\n@sealed"),
            [
                _define("ns/Type_.2.0.uavcan", "@sealed"),
            ],
        )

    with raises(_error.InvalidDefinitionError, match="(?i).*Bit length cannot exceed.*"):
        _parse_definition(
            _define("vendor/types/A.1.0.uavcan", "int128 field\n@sealed"),
            [
                _define("ns/Type_.2.0.uavcan", "@sealed"),
                _define("ns/Type_.1.1.uavcan", "@sealed"),
            ],
        )

    with raises(_error.InvalidDefinitionError, match="(?i).*type.*"):
        _parse_definition(
            _define("vendor/invalid_constant_value/A.1.0.uavcan", "ns.Type_.1.1 VALUE = 123\n@sealed"),
            [
                _define("ns/Type_.2.0.uavcan", "@sealed"),
                _define("ns/Type_.1.1.uavcan", "@sealed"),
            ],
        )

    with raises(_data_type_builder.UndefinedDataTypeError):
        defs = [
            _define("vendor/circular_dependency/A.1.0.uavcan", "B.1.0 b\n@sealed"),
            _define("vendor/circular_dependency/B.1.0.uavcan", "A.1.0 b\n@sealed"),
        ]
        _parse_definition(defs[0], defs)

    with raises(_error.InvalidDefinitionError, match="(?i).*union directive.*"):
        _parse_definition(
            _define("vendor/misplaced_directive/A.1.0.uavcan", "ns.Type_.2.0 field\n@union\n@sealed"),
            [
                _define("ns/Type_.2.0.uavcan", "@sealed"),
            ],
        )

    with raises(_error.InvalidDefinitionError, match="(?i).*deprecated directive.*"):
        _parse_definition(
            _define("vendor/misplaced_directive/A.1.0.uavcan", "ns.Type_.2.0 field\n@deprecated\n@sealed"),
            [
                _define("ns/Type_.2.0.uavcan", "@sealed"),
            ],
        )

    with raises(_error.InvalidDefinitionError, match="(?i).*deprecated directive.*"):
        _parse_definition(
            _define(
                "vendor/misplaced_directive/A.1.0.uavcan", "ns.Type_.2.0 field\n@sealed\n---\n@deprecated\n@sealed"
            ),
            [
                _define("ns/Type_.2.0.uavcan", "@sealed"),
            ],
        )

    try:
        standalone(
            "vendor/types/A.1.0.uavcan",
            dedent(
                """
                   int8 a  # Comment
                   # Empty
                   @assert false  # Will error here, line number 4
                   # Blank
                   @sealed
                   """
            ),
        )
    except _error.FrontendError as ex:
        assert ex.path and str(ex.path).endswith(os.path.join("vendor", "types", "A.1.0.uavcan"))
        assert ex.line and ex.line == 4
    else:  # pragma: no cover
        assert False

    standalone("vendor/types/1.A.1.0.uavcan", "@sealed", allow_unregulated=True)
    with raises(_data_type_builder.UnregulatedFixedPortIDError, match=r".*allow_unregulated_fixed_port_id.*"):
        standalone("vendor/types/1.A.1.0.uavcan", "@sealed")

    standalone("vendor/types/1.A.1.0.uavcan", "@sealed\n---\n@sealed", allow_unregulated=True)
    with raises(_data_type_builder.UnregulatedFixedPortIDError, match=r".*allow_unregulated_fixed_port_id.*"):
        standalone("vendor/types/1.A.1.0.uavcan", "@sealed\n---\n@sealed")

    with raises(_error.InvalidDefinitionError, match="(?i).*seal.*"):
        standalone(
            "vendor/sealing/A.1.0.uavcan",
            dedent(
                """
                   int8 a
                   @extent 128
                   @sealed
                   """
            ),
        )

    with raises(_error.InvalidDefinitionError, match="(?i).*extent.*"):
        standalone(
            "vendor/sealing/A.1.0.uavcan",
            dedent(
                """
                   int8 a
                   @sealed
                   @extent 128
                   """
            ),
        )

    with raises(_error.InvalidDefinitionError, match="(?i).*sealed.*expression.*"):
        standalone(
            "vendor/sealing/A.1.0.uavcan",
            dedent(
                """
                   int8 a
                   @sealed 12345678
                   """
            ),
        )

    with raises(_error.InvalidDefinitionError, match="(?i).*extent.*expression.*"):
        standalone(
            "vendor/sealing/A.1.0.uavcan",
            dedent(
                """
                   int8 a
                   @extent
                   """
            ),
        )

    with raises(_error.InvalidDefinitionError, match="(?i).*extent.*"):
        standalone(
            "vendor/sealing/A.1.0.uavcan",
            dedent(
                """
                   int16 a
                   @extent 8  # Too small
                   """
            ),
        )

    with raises(_error.InvalidDefinitionError, match="(?i).*extent.*"):
        standalone(
            "vendor/sealing/A.1.0.uavcan",
            dedent(
                """
                   int16 a
                   @extent {16}  # Wrong type
                   """
            ),
        )

    with raises(_error.InvalidDefinitionError, match="(?i).*extent.*"):
        standalone(
            "vendor/sealing/A.1.0.uavcan",
            dedent(
                """
                   int16 a
                   @extent 64
                   int8 b
                   """
            ),
        )

    with raises(_error.InvalidDefinitionError, match="(?i).*extent.*"):  # Neither extent nor sealed are specified.
        standalone(
            "vendor/sealing/A.1.0.uavcan",
            dedent(
                """
                   int16 a
                   int8 b
                   """
            ),
        )


@_in_n_out
def _unittest_print() -> None:
    printed_items = None  # type: Optional[Tuple[int, str]]

    def print_handler(line_number: int, text: str) -> None:
        nonlocal printed_items
        printed_items = line_number, text

    _define(
        "ns/A.1.0.uavcan",
        "# line number 1\n" "# line number 2\n" "@print 2 + 2 == 4   # line number 3\n" "# line number 4\n" "@sealed\n",
    ).read([], print_handler, False)

    assert printed_items
    assert printed_items[0] == 3
    assert printed_items[1] == "true"

    _define("ns/B.1.0.uavcan", "@print false\n@sealed").read([], print_handler, False)
    assert printed_items
    assert printed_items[0] == 1
    assert printed_items[1] == "false"

    _define(
        "ns/Offset.1.0.uavcan", "@print _offset_    # Not recorded\n" "uint8 a\n" "@print _offset_\n" "@extent 800\n"
    ).read([], print_handler, False)
    assert printed_items
    assert printed_items[0] == 3
    assert printed_items[1] == "{8}"


# noinspection PyProtectedMember
@_in_n_out
def _unittest_assert() -> None:
    from pytest import raises

    _parse_definition(
        _define(
            "ns/A.1.0.uavcan",
            dedent(
                """
            @assert _offset_ == {0}
            @assert _offset_.min == _offset_.max
            Array.1.0[2] bar
            @assert _offset_ == {16, 24, 32, 40, 48}
            @assert _offset_.min == 16
            @assert _offset_.max == 48
            @assert _offset_ % 8 == {0}
            @assert _offset_ % 10 == {6, 4, 2, 0, 8}
            @assert _offset_ * 2 == {32, 48, 64, 80, 96}
            @assert 2 * _offset_ == {32, 48, 64, 80, 96}
            @assert _offset_ / 4 == {4, 6, 8, 10, 12}
            @assert _offset_ - 4 == {12, 20, 28, 36, 44}
            @assert _offset_ + 4 == {20, 28, 36, 44, 52}
            uint64 big
            @assert _offset_ - 64 == {16, 24, 32, 40, 48}
            @assert _offset_.min == 80
            @assert _offset_.max == 112
            @assert _offset_.max <= 112
            @assert _offset_.max < 113
            @assert _offset_ == _offset_
            @assert truncated uint64._bit_length_ == {64}
            @assert uint64._bit_length_ == {64}
            @assert Array.1.0._bit_length_.max == 8 + 8 + 8
            @assert Array.1.0._extent_ == 8 + 8 + 8
            @assert Array.1.0._extent_ == Array.1.0._bit_length_.max
            @sealed
            """
            ),
        ),
        [_define("ns/Array.1.0.uavcan", "uint8[<=2] foo\n@sealed")],
    )

    with raises(_error.InvalidDefinitionError, match="(?i).*operator is not defined.*"):
        _parse_definition(
            _define(
                "ns/C.1.0.uavcan",
                dedent(
                    """
                uint64 big
                @assert _offset_ == 64
                @sealed
                """
                ),
            ),
            [],
        )

    with raises(_expression.UndefinedAttributeError):
        _parse_definition(
            _define("ns/C.1.0.uavcan", "@print Service.1.0._bit_length_"),
            [_define("ns/Service.1.0.uavcan", "uint8 a\n@sealed\n---\nuint16 b\n@sealed")],
        )

    with raises(_expression.UndefinedAttributeError):
        _parse_definition(_define("ns/C.1.0.uavcan", """uint64 LENGTH = uint64.nonexistent_attribute\n@extent 0"""), [])

    with raises(_error.InvalidDefinitionError, match="(?i).*void.*"):
        _parse_definition(_define("ns/C.1.0.uavcan", "void2 name\n@sealed"), [])

    with raises(_serializable._attribute.InvalidConstantValueError):
        _parse_definition(_define("ns/C.1.0.uavcan", "int8 name = true\n@sealed"), [])

    with raises(_error.InvalidDefinitionError, match=".*value.*"):
        _parse_definition(_define("ns/C.1.0.uavcan", "int8 name = {1, 2, 3}\n@sealed"), [])

    _parse_definition(
        _define(
            "ns/D.1.0.uavcan",
            dedent(
                """
            @union
            float32 a
            uint64 b
            @assert _offset_ == {40, 72}
            @sealed
            """
            ),
        ),
        [],
    )

    _parse_definition(
        _define(
            "ns/E.1.0.uavcan",
            dedent(
                """
            @union
            uint8 A = 0
            float32 a
            uint8 B = 1
            uint64 b
            uint8 C = 2
            @assert _offset_ == {40, 72}
            uint8 D = 3
            @sealed
            """
            ),
        ),
        [],
    )

    with raises(_error.InvalidDefinitionError):
        _parse_definition(
            _define(
                "ns/F.1.0.uavcan",
                dedent(
                    """
                @union
                @assert _offset_.min == 33
                float32 a
                uint64 b
                @assert _offset_ == {40, 72}
                @sealed
                """
                ),
            ),
            [],
        )

    with raises(_data_type_builder.AssertionCheckFailureError):
        _parse_definition(
            _define(
                "ns/G.1.0.uavcan",
                dedent(
                    """
                float32 a
                @assert _offset_.min == 8
                @sealed
                """
                ),
            ),
            [],
        )

    with raises(_error.InvalidDefinitionError, match="(?i).*yield a boolean.*"):
        _parse_definition(
            _define(
                "ns/H.1.0.uavcan",
                dedent(
                    """
                float32 a
                @assert _offset_.min
                @sealed
                """
                ),
            ),
            [],
        )

    # Extent verification
    _parse_definition(
        _define(
            "ns/I.1.0.uavcan",
            dedent(
                """
            @assert J.1.0._extent_ == 64
            @assert J.1.0._bit_length_ == {0, 1, 2, 3, 4, 5, 6, 7, 8} * 8 + 32
            @assert K.1.0._extent_ == 8
            @assert K.1.0._bit_length_ == {8}
            @sealed
            """
            ),
        ),
        [
            _define("ns/J.1.0.uavcan", "uint8 foo\n@extent 64"),
            _define("ns/K.1.0.uavcan", "uint8 foo\n@sealed"),
        ],
    )

    # Alignment
    _parse_definition(
        _define(
            "ns/L.1.0.uavcan",
            dedent(
                """
            @assert _offset_ == {0}
            uint3 a
            @assert _offset_ == {3}
            N.1.0 nothing
            @print _offset_
            @assert _offset_ == {8}   # Aligned!
            uint5 b
            @assert _offset_ == {13}
            N.1.0[3] array_of_nothing
            @assert _offset_ == {16}  # Aligned!
            bool c
            @assert _offset_ == {17}
            M.1.0 variable
            @assert _offset_ == 32 + {24, 32, 40}  # Aligned; variability due to extensibility (non-sealing)
            @sealed
            """
            ),
        ),
        [
            _define("ns/M.1.0.uavcan", "@extent 16"),
            _define("ns/N.1.0.uavcan", "@sealed"),
        ],
    )


def _unittest_parse_namespace() -> None:
    from pytest import raises

    directory = tempfile.TemporaryDirectory()

    print_output = None  # type: Optional[Tuple[str, int, str]]

    def print_handler(d: Path, line: int, text: str) -> None:
        nonlocal print_output
        print_output = str(d), line, text

    # noinspection PyShadowingNames
    def _define(rel_path: str, text: str) -> None:
        path = Path(directory.name, rel_path)
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, "w") as f:
            f.write(text)

    # Empty namespace.
    assert [] == _namespace.read_namespace(directory.name)

    _define(
        "zubax/First.1.0.uavcan",
        dedent(
            """
        uint8[<256] a
        @assert _offset_.min == 8
        @assert _offset_.max == 2048
        @sealed
        """
        ),
    )

    _define(
        "zubax/7001.Message.1.0.uavcan",
        dedent(
            """
        zubax.First.1.0[<=2] a
        @assert _offset_.min == 8
        @assert _offset_.max == 4104
        @extent _offset_.max * 8
        """
        ),
    )

    _define(
        "zubax/nested/300.Spartans.30.0.uavcan",
        dedent(
            """
        @deprecated
        @union
        float16 small
        float32 just_right
        float64 woah
        @extent _offset_.max * 8
        ---
        @print _offset_     # Will print zero {0}
        @sealed
        """
        ),
    )

    _define("zubax/nested/300.Spartans.30.0.txt", "completely unrelated stuff")
    _define("zubax/300.Spartans.30.0", "completely unrelated stuff")

    parsed = _namespace.read_namespace(
        Path(directory.name, "zubax"),
        [Path(directory.name, "zubax", ".")],  # Intentional duplicate
        print_handler,
    )
    print(parsed)
    assert len(parsed) == 3
    assert "zubax.First" in [x.full_name for x in parsed]
    assert "zubax.Message" in [x.full_name for x in parsed]
    assert "zubax.nested.Spartans" in [x.full_name for x in parsed]

    # try again with minimal arguments to read_namespace
    parsed_minimal_args = _namespace.read_namespace(Path(directory.name, "zubax"))
    assert len(parsed_minimal_args) == 3

    _define(
        "zubax/colliding/300.Iceberg.30.0.uavcan",
        dedent(
            """
        @extent 1024
        ---
        @extent 1024
        """
        ),
    )

    with raises(_namespace.FixedPortIDCollisionError):
        _namespace.read_namespace(Path(directory.name, "zubax"), [], print_handler)

    with raises(TypeError):  # Invalid usage: expected path-like object, not bytes.
        _namespace.read_namespace(Path(directory.name, "zubax"), b"/my/path")  # type: ignore

    with raises(TypeError):  # Invalid usage: expected path-like object, not bytes.
        # noinspection PyTypeChecker
        _namespace.read_namespace(Path(directory.name, "zubax"), [b"/my/path"])  # type: ignore

    assert print_output is not None
    assert "300.Spartans" in print_output[0]
    assert print_output[1] == 9
    assert print_output[2] == "{0}"

    _define(
        "zubax/colliding/iceberg/300.Ice.30.0.uavcan",
        dedent(
            """
        @sealed
        ---
        @sealed
        """
        ),
    )
    with raises(_namespace.DataTypeNameCollisionError):
        _namespace.read_namespace(
            Path(directory.name, "zubax"),
            [
                Path(directory.name, "zubax"),
            ],
        )

    # Do again to test single lookup-directory override
    with raises(_namespace.DataTypeNameCollisionError):
        _namespace.read_namespace(Path(directory.name, "zubax"), Path(directory.name, "zubax"))

    try:
        os.unlink(Path(directory.name, "zubax/colliding/iceberg/300.Ice.30.0.uavcan"))
        _define(
            "zubax/COLLIDING/300.Iceberg.30.0.uavcan",
            dedent(
                """
            @extent 1024
            ---
            @extent 1024
            """
            ),
        )
        with raises(_namespace.DataTypeNameCollisionError, match=".*letter case.*"):
            _namespace.read_namespace(
                Path(directory.name, "zubax"),
                [
                    Path(directory.name, "zubax"),
                ],
            )
    except _namespace.FixedPortIDCollisionError:  # pragma: no cover
        pass  # We're running on a platform where paths are not case-sensitive.


def _unittest_parse_namespace_versioning() -> None:
    from pytest import raises
    import glob

    directory = tempfile.TemporaryDirectory()

    # noinspection PyShadowingNames
    def _define(rel_path: str, text: str) -> None:
        path = Path(directory.name, rel_path)
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, "w") as f:
            f.write(text)

    def _undefine_glob(rel_path_glob: str) -> None:
        for g in glob.glob(str(Path(directory.name, rel_path_glob))):
            os.remove(g)

    _define(
        "ns/Spartans.30.0.uavcan",
        dedent(
            """
        @deprecated
        @union
        float16 small
        float32 just_right
        float64 woah
        @extent 1024
        ---
        @extent 1024
        """
        ),
    )

    _define(
        "ns/Spartans.30.1.uavcan",
        dedent(
            """
        @deprecated
        @union
        uint16 small
        int32 just_right
        float64[1] woah
        @extent 1024
        ---
        @extent 1024
        """
        ),
    )

    parsed = _namespace.read_namespace(Path(directory.name, "ns"), [])
    print(parsed)
    assert len(parsed) == 2

    _define(
        "ns/Spartans.30.2.uavcan",
        dedent(
            """
        @deprecated
        @union
        uint16 small
        int32 just_right
        float64[1] woah
        @extent 1024
        """
        ),
    )

    with raises(_namespace.VersionsOfDifferentKindError):
        _namespace.read_namespace(Path(directory.name, "ns"), [])

    _undefine_glob("ns/Spartans.30.[01].uavcan")

    _define(
        "ns/Spartans.30.0.uavcan",
        dedent(
            """
        @deprecated
        @union
        uint16 small
        float32 just_right
        float64[1] woah
        @extent 1024
        """
        ),
    )

    parsed = _namespace.read_namespace(Path(directory.name, "ns"), [])
    print(parsed)
    assert len(parsed) == 2

    _define(
        "ns/Spartans.30.1.uavcan",
        dedent(
            """
        @deprecated
        @union
        uint16 small
        float32 just_right
        int64 woah
        @extent 1024
        """
        ),
    )

    _define(
        "ns/6700.Spartans.30.2.uavcan",
        dedent(
            """
        @deprecated
        @union
        uint16 small
        int32 just_right
        float64[1] woah
        @extent 1024
        """
        ),
    )

    with raises(_namespace.MultipleDefinitionsUnderSameVersionError):
        _namespace.read_namespace(Path(directory.name, "ns"), [])

    _undefine_glob("ns/Spartans.30.2.uavcan")

    parsed = _namespace.read_namespace(Path(directory.name, "ns"), [])
    assert len(parsed) == 3

    _undefine_glob("ns/Spartans.30.0.uavcan")
    _define(
        "ns/6700.Spartans.30.0.uavcan",
        dedent(
            """
        @deprecated
        @union
        uint16 small
        float32 just_right
        float64[1] woah
        @extent 1024
        """
        ),
    )

    with raises(_namespace.MinorVersionFixedPortIDError):
        _namespace.read_namespace(Path(directory.name, "ns"), [])

    _undefine_glob("ns/Spartans.30.1.uavcan")
    _define(
        "ns/6700.Spartans.30.1.uavcan",
        dedent(
            """
        @deprecated
        @union
        uint16 small
        float32 just_right
        float64[1] woah
        @extent 1024
        """
        ),
    )

    parsed = _namespace.read_namespace(Path(directory.name, "ns"), [])
    assert len(parsed) == 3

    _undefine_glob("ns/6700.Spartans.30.1.uavcan")
    _define(
        "ns/6701.Spartans.30.1.uavcan",
        dedent(
            """
        @deprecated
        @union
        uint16 small
        float32 just_right
        float64[1] woah
        @extent 1024
        """
        ),
    )

    with raises(_namespace.MinorVersionFixedPortIDError):
        _namespace.read_namespace(Path(directory.name, "ns"), [])

    # Adding new major version under the same FPID
    _undefine_glob("ns/6701.Spartans.30.1.uavcan")
    _define(
        "ns/6700.Spartans.31.0.uavcan",
        dedent(
            """
        @deprecated
        @union
        uint16 small
        float32 just_right
        float64[1] woah
        @extent 1024
        """
        ),
    )

    with raises(_namespace.FixedPortIDCollisionError):
        _namespace.read_namespace(Path(directory.name, "ns"), [])

    # Major version zero allows us to re-use the same FPID under a different (non-zero) major version
    _undefine_glob("ns/6700.Spartans.31.0.uavcan")
    _define(
        "ns/6700.Spartans.0.1.uavcan",
        dedent(
            """
        @deprecated
        @union
        uint16 small
        float32 just_right
        float64[1] woah
        @extent 1024
        """
        ),
    )

    # These are needed to ensure full branch coverage, see the checking code.
    _define("ns/Empty.1.0.uavcan", "@extent 0")
    _define("ns/Empty.1.1.uavcan", "@extent 0")
    _define("ns/Empty.2.0.uavcan", "@extent 0")
    _define("ns/6800.Empty.3.0.uavcan", "@extent 0")
    _define("ns/6801.Empty.4.0.uavcan", "@extent 0")

    parsed = _namespace.read_namespace(Path(directory.name, "ns"), [])  # no error
    assert len(parsed) == 8

    # Check ordering - the definitions must be sorted properly by name (lexicographically) and version (newest first).
    assert list(map(str, parsed)) == [
        "ns.Empty.4.0",  # Newest first
        "ns.Empty.3.0",
        "ns.Empty.2.0",
        "ns.Empty.1.1",
        "ns.Empty.1.0",
        "ns.Spartans.30.2",  # Newest first
        "ns.Spartans.30.0",
        "ns.Spartans.0.1",
    ]

    # Extent consistency -- non-service type
    _define("ns/Consistency.1.0.uavcan", "uint8 a\n@extent 128")
    _define("ns/Consistency.1.1.uavcan", "uint8 a\nuint8 b\n@extent 128")
    parsed = _namespace.read_namespace(Path(directory.name, "ns"), [])  # no error
    assert len(parsed) == 10
    _define("ns/Consistency.1.2.uavcan", "uint8 a\nuint8 b\nuint8 c\n@extent 256")
    with raises(
        _namespace.ExtentConsistencyError, match=r"(?i).*extent of ns\.Consistency\.1\.2 is 256 bits.*"
    ) as ei_extent:
        _namespace.read_namespace(Path(directory.name, "ns"), [])
    print(ei_extent.value)
    assert ei_extent.value.path and "Consistency.1" in str(ei_extent.value.path)
    _undefine_glob("ns/Consistency*")

    # Extent consistency -- non-service type, zero major version
    _define("ns/Consistency.0.1.uavcan", "uint8 a\n@extent 128")
    _define("ns/Consistency.0.2.uavcan", "uint8 a\nuint8 b\n@extent 128")
    parsed = _namespace.read_namespace(Path(directory.name, "ns"), [])  # no error
    assert len(parsed) == 10
    _define("ns/Consistency.0.3.uavcan", "uint8 a\nuint8 b\nuint8 c\n@extent 256")  # no error
    _namespace.read_namespace(Path(directory.name, "ns"), [])
    _undefine_glob("ns/Consistency*")

    # Extent consistency -- request
    _define(
        "ns/Consistency.1.0.uavcan",
        dedent(
            """
            uint8 a
            @extent 128
            ---
            uint8 a
            @extent 128
            """
        ),
    )
    _define(
        "ns/Consistency.1.1.uavcan",
        dedent(
            """
            uint8 a
            uint8 b
            @extent 128
            ---
            uint8 a
            @extent 128
            """
        ),
    )
    parsed = _namespace.read_namespace(Path(directory.name, "ns"), [])  # no error
    assert len(parsed) == 10
    _define(
        "ns/Consistency.1.2.uavcan",
        dedent(
            """
            uint8 a
            uint8 b
            @extent 256
            ---
            uint8 a
            @extent 128
            """
        ),
    )
    with raises(
        _namespace.ExtentConsistencyError, match=r"(?i).*extent of ns\.Consistency.* is 256 bits.*"
    ) as ei_extent:
        _namespace.read_namespace(Path(directory.name, "ns"), [])
    print(ei_extent.value)
    assert ei_extent.value.path and "Consistency.1" in str(ei_extent.value.path)
    _undefine_glob("ns/Consistency*")

    # Extent consistency -- response
    _define(
        "ns/Consistency.1.0.uavcan",
        dedent(
            """
            uint8 a
            @extent 128
            ---
            uint8 a
            @extent 128
            """
        ),
    )
    _define(
        "ns/Consistency.1.1.uavcan",
        dedent(
            """
            uint8 a
            @extent 128
            ---
            uint8 a
            uint8 b
            @extent 128
            """
        ),
    )
    parsed = _namespace.read_namespace(Path(directory.name, "ns"), [])  # no error
    assert len(parsed) == 10
    _define(
        "ns/Consistency.1.2.uavcan",
        dedent(
            """
            uint8 a
            @extent 128
            ---
            uint8 a
            @extent 256
            """
        ),
    )
    with raises(
        _namespace.ExtentConsistencyError, match=r"(?i).*extent of ns\.Consistency.* is 256 bits.*"
    ) as ei_extent:
        _namespace.read_namespace(Path(directory.name, "ns"), [])
    print(ei_extent.value)
    assert ei_extent.value.path and "Consistency.1" in str(ei_extent.value.path)
    _undefine_glob("ns/Consistency*")

    # Sealing consistency -- non-service type
    _define("ns/Consistency.1.0.uavcan", "uint64 a\n@extent 64")
    _define("ns/Consistency.1.1.uavcan", "uint64 a\n@extent 64")
    parsed = _namespace.read_namespace(Path(directory.name, "ns"), [])  # no error
    assert len(parsed) == 10
    _define("ns/Consistency.1.2.uavcan", "uint64 a\n@sealed")
    with raises(_namespace.SealingConsistencyError, match=r"(?i).*ns\.Consistency\.1\.2 is sealed.*") as ei_sealing:
        _namespace.read_namespace(Path(directory.name, "ns"), [])
    print(ei_sealing.value)
    assert ei_sealing.value.path and "Consistency.1" in str(ei_sealing.value.path)
    _undefine_glob("ns/Consistency*")

    # Sealing consistency -- request
    _define(
        "ns/Consistency.1.0.uavcan",
        dedent(
            """
            uint64 a
            @extent 64
            ---
            uint64 a
            @extent 64
            """
        ),
    )
    _define(
        "ns/Consistency.1.1.uavcan",
        dedent(
            """
            uint64 a
            @extent 64
            ---
            uint64 a
            @extent 64
            """
        ),
    )
    parsed = _namespace.read_namespace(Path(directory.name, "ns"), [])  # no error
    assert len(parsed) == 10
    _define(
        "ns/Consistency.1.2.uavcan",
        dedent(
            """
            uint64 a
            @sealed
            ---
            uint64 a
            @extent 64
            """
        ),
    )
    with raises(_namespace.SealingConsistencyError, match=r"(?i).*ns\.Consistency.* is sealed.*") as ei_sealing:
        _namespace.read_namespace(Path(directory.name, "ns"), [])
    print(ei_sealing.value)
    assert ei_sealing.value.path and "Consistency.1" in str(ei_sealing.value.path)
    _undefine_glob("ns/Consistency*")

    # Sealing consistency -- response
    _define(
        "ns/Consistency.1.0.uavcan",
        dedent(
            """
            uint64 a
            @extent 64
            ---
            uint64 a
            @extent 64
            """
        ),
    )
    _define(
        "ns/Consistency.1.1.uavcan",
        dedent(
            """
            uint64 a
            @extent 64
            ---
            uint64 a
            @extent 64
            """
        ),
    )
    parsed = _namespace.read_namespace(Path(directory.name, "ns"), [])  # no error
    assert len(parsed) == 10
    _define(
        "ns/Consistency.1.2.uavcan",
        dedent(
            """
            uint64 a
            @extent 64
            ---
            uint64 a
            @sealed
            """
        ),
    )
    with raises(_namespace.SealingConsistencyError, match=r"(?i).*ns\.Consistency.* is sealed.*") as ei_sealing:
        _namespace.read_namespace(Path(directory.name, "ns"), [])
    print(ei_sealing.value)
    assert ei_sealing.value.path and "Consistency.1" in str(ei_sealing.value.path)
    _undefine_glob("ns/Consistency*")


def _unittest_parse_namespace_faults() -> None:
    try:
        _namespace.read_namespace("/foo/bar/baz", ["/bat/wot", "/foo/bar/baz/bad"])
    except _namespace.NestedRootNamespaceError as ex:
        print(ex)
    else:  # pragma: no cover
        assert False

    try:
        _namespace.read_namespace("/foo/bar/baz", ["/foo/bar/zoo", "/foo/bar/doo/roo/BAZ"])  # Notice the letter case
    except _namespace.RootNamespaceNameCollisionError as ex:
        print(ex)
    else:  # pragma: no cover
        assert False
    try:
        _namespace.read_namespace("/foo/bar/baz", ["/foo/bar/zoo", "/foo/bar/doo/roo/zoo", "/foo/bar/doo/roo/baz"])
    except _namespace.RootNamespaceNameCollisionError as ex:
        print(ex)
    else:  # pragma: no cover
        assert False


@_in_n_out
def _unittest_inconsistent_deprecation() -> None:
    from pytest import raises

    _parse_definition(
        _define("ns/A.1.0.uavcan", "@sealed"),
        [
            _define(
                "ns/B.1.0.uavcan",
                dedent(
                    """
                    @deprecated
                    A.1.0 a
                    @sealed
                    """
                ),
            )
        ],
    )

    with raises(_error.InvalidDefinitionError, match="(?i).*depend.*deprecated.*"):
        _parse_definition(
            _define(
                "ns/C.1.0.uavcan",
                dedent(
                    """
                X.1.0 b
                @sealed
                """
                ),
            ),
            [_define("ns/X.1.0.uavcan", "@deprecated\n@sealed")],
        )

    _parse_definition(
        _define(
            "ns/D.1.0.uavcan",
            dedent(
                """
                @deprecated
                X.1.0 b
                @sealed
                """
            ),
        ),
        [_define("ns/X.1.0.uavcan", "@deprecated\n@sealed")],
    )


@_in_n_out
def _unittest_repeated_directives() -> None:
    from pytest import raises

    _parse_definition(
        _define(
            "ns/A.1.0.uavcan",
            dedent(
                """
                @union
                @deprecated
                int8 a
                float16 b
                @sealed
                """
            ),
        ),
        [],
    )

    with raises(_error.InvalidDefinitionError, match="(?i).*deprecated.*"):
        _parse_definition(
            _define(
                "ns/A.1.0.uavcan",
                dedent(
                    """
                    @deprecated
                    @deprecated
                    @sealed
                    """
                ),
            ),
            [],
        )

    with raises(_error.InvalidDefinitionError, match="(?i).*deprecated.*"):
        _parse_definition(
            _define(
                "ns/A.1.0.uavcan",
                dedent(
                    """
                    @deprecated
                    @sealed
                    ---
                    @deprecated
                    @sealed
                    """
                ),
            ),
            [],
        )

    _parse_definition(
        _define(
            "ns/A.1.0.uavcan",
            dedent(
                """
                @union
                int8 a
                float16 b
                @sealed
                ---
                @union
                int8 a
                float16 b
                @sealed
                """
            ),
        ),
        [],
    )

    with raises(_error.InvalidDefinitionError, match="(?i).*union.*"):
        _parse_definition(
            _define(
                "ns/A.1.0.uavcan",
                dedent(
                    """
                    @union
                    @union
                    int8 a
                    float16 b
                    @sealed
                    """
                ),
            ),
            [],
        )

    with raises(_error.InvalidDefinitionError, match="(?i).*sealed.*"):
        _parse_definition(
            _define(
                "ns/A.1.0.uavcan",
                dedent(
                    """
                    @sealed
                    @sealed
                    int8 a
                    float16 b
                    @sealed
                    """
                ),
            ),
            [],
        )

    with raises(_error.InvalidDefinitionError, match="(?i).*extent.*already set.*"):
        _parse_definition(
            _define(
                "ns/A.1.0.uavcan",
                dedent(
                    """
                    int8 a
                    float16 b
                    @extent 256
                    @extent 800
                    @sealed
                    """
                ),
            ),
            [],
        )


@_in_n_out
def _unittest_dsdl_parser_basics() -> None:
    # This is how you can run one test only for development needs:
    #   pytest pydsdl -k _unittest_dsdl_parser_basics --capture=no
    # noinspection SpellCheckingInspection
    _parse_definition(
        _define(
            "ns/A.1.0.uavcan",
            dedent(
                r"""
                @deprecated
                void16
                int8           [<=123+456] array_inclusive
                truncated uint8[< 123+456] array_exclusive
                saturated int8 [  123+456] array_fixed
                #ns.Bar.1.23 field
                float64 a = +10 * (-2 / -3) / 4 % 5
                bool    b = !true
                float32 c = (123456 + 0x_ab_cd_ef) / 0b1111_1111 ** 2 - 0o123_456 * 2.7
                @print "Hello\r\nworld!"
                @print
                @assert true
                @assert ns.Foo.1.0.THE_CONSTANT == 42
                @assert ns.Bar.1.23.B == ns.Bar.1.23.A + 1
                @extent 32 * 1024 * 8
                """
            ),
        ),
        [
            _define("ns/Foo.1.0.uavcan", "int8 THE_CONSTANT = 42\n@extent 1024"),
            _define("ns/Bar.1.23.uavcan", "int8 the_field\nint8 A = 0xA\nint8 B = 0xB\n@extent 1024"),
        ],
    )


@_in_n_out
def _unittest_dsdl_parser_expressions() -> None:
    from pytest import raises

    def throws(definition: str, exc: typing.Type[Exception] = _expression.InvalidOperandError) -> None:
        with raises(exc):
            _parse_definition(_define("ns/Throws.0.1.uavcan", dedent(definition + "\n@sealed")), [])

    throws("bool R = true && 0")
    throws("bool R = true || 0")
    throws("bool R = 0 || true")
    throws("bool R = 0 == true")
    throws("bool R = {0} & true")
    throws("bool R = true ^ {0}")
    throws("bool R = 0 ^ true")
    throws("int8 R = 1 / 0")
    throws('bool R = "S" == 0')
    throws("bool R = {0} != {}")
    throws('bool R = {0, true, "S"}')
    throws('bool R = {0} == {"s"}')
    throws('bool R = {0} <= "s"')
    throws('bool R = {0} >= "s"')
    throws('bool R = {0} > "s"')
    throws('bool R = {0} < "s"')
    throws('bool R = {0} | "s"')
    throws('bool R = {0} & "s"')
    throws('bool R = {0} ^ "s"')
    throws("bool R = {0}.nonexistent_attribute")
    throws("bool R = {0} / {1}")
    throws("bool R = !1")
    throws("bool R = +true")
    throws('bool R = -"1"')
    throws("bool R = true | false")
    throws("bool R = true & false")
    throws('bool R = true + "0"')
    throws('bool R = true - "0"')
    throws('bool R = true * "0"')
    throws('bool R = true / "0"')
    throws('bool R = true % "0"')
    throws('bool R = true ** "0"')

    _parse_definition(
        _define(
            "ns/A.1.0.uavcan",
            dedent(
                r"""
                float64 PI = 3.141592653589793
                float64 E  = 2.718281828459045
                @assert (PI ** E > 22.4) && (PI ** E < 22.5)
                @assert 'moments of eternity'     != "strangers stealing someone else's dreams"  # I've seen it all
                @assert 'hunting for the mystery' != 'running for your life in times like these' # I've seen it all
                @assert "I remember the time once in a life" != 'oh baby'  # got you here in my head, here in my head
                @assert false == ('oh' == 'maybe')
                @assert true
                @assert 1 == 2 - 1
                @assert -10 == +20 / -2
                @assert {10, 15, 20} % 5 == {0}
                @assert {10, 15, 20} % 5 == {1, 2, 3} * 0
                @assert {10, 15, 20} / 5 == {2, 3, 4} * 10 / 5 / 2
                @assert {1} < {1, 2}
                @assert {1} <= {1}
                @assert {1} != {1, 2}
                @assert {1} >= {1}
                @assert {1, 2} > {1, 2} == false
                @assert {1, 2, 3} > {1, 2}
                @assert {1, 5/2} == {-5/-2, 2.5, 1}
                @assert {1, 2, 3} == {1} | {2} | {3} | {1}
                @assert {1, 2, 3} == {1, 2, 3, 4, 5} & {1, 2, 3, 8, 9}
                @assert {4, 5, 8, 9} == {1, 2, 3, 4, 5} ^ {1, 2, 3, 8, 9}
                @assert 1 - {1, 2, 3} == {0, -1, -2}
                @assert 1 / {1, 2, 3} == {1, 1/2, 1/3}
                @assert 8 % {1, 2, 3} == {0, 2}
                @assert 2 ** {1, 2, 3} == {2, 4, 8}
                @assert {1, 2, 3} ** 2 == {1, 4, 9}
                @assert "Hello" + ' ' + 'world' == 'Hello world'
                @assert 'Hello'+' '+'world' != ''
                @assert '\u00e9' == '\u0065\u0301'  # e with accent
                @assert '\ufb03' != 'ffi'           # ffi ligature is not decomposed by NFC
                @assert true != ('A' == "a")
                @assert true && true
                @assert ! (true && false)
                @assert true||false
                @assert !false
                @assert ! 5 < 3
                @assert 4 | 2 == 6
                @assert 4 & 2 == 0
                @assert 3 & 2 == 2
                @assert 0xFF_00 & 0x00_FF == 0x0000
                @assert 0xFF_00 | 0x00_FF == 0xFFFF
                @assert 0xFF_00 ^ 0x0F_FF == 0xF0FF
                @sealed
                """
            ),
        ),
        [],
    )


@_in_n_out
def _unittest_pickle() -> None:
    import pickle

    p = _parse_definition(
        _define(
            "ns/A.1.0.uavcan",
            dedent(
                r"""
                float64 PI = 3.141592653589793
                float64 big_pi
                @sealed
                ---
                float16 small_pi
                @extent 1024 * 8
                """
            ),
        ),
        [],
    )
    assert isinstance(p, _serializable.ServiceType)
    assert p.request_type.has_parent_service
    assert p.response_type.has_parent_service
    assert not p.has_parent_service

    pp = pickle.loads(pickle.dumps(p))
    assert isinstance(pp, _serializable.ServiceType)
    assert pp.request_type.has_parent_service
    assert pp.response_type.has_parent_service
    assert not pp.has_parent_service
    assert str(pp) == str(p)
    assert repr(pp) == repr(p)


def _collect_descendants(cls: typing.Type[object]) -> typing.Iterable[typing.Type[object]]:
    # noinspection PyArgumentList
    for t in cls.__subclasses__():
        yield t
        yield from _collect_descendants(t)


def _unittest_collect_descendants() -> None:  # Unit test for my unit test.
    class A:
        pass

    class B(A):
        pass

    class C(B):
        pass

    class D(A):
        pass

    assert set(_collect_descendants(A)) == {B, C, D}
    assert set(_collect_descendants(D)) == set()
    assert bool in set(_collect_descendants(int))


def _unittest_public_api() -> None:
    import pydsdl

    # Ensure that all descendants of the specified classes are exported from the library.
    # If this test fails, you probably forgot to update __init__.py.
    public_roots = [
        _serializable.SerializableType,
        _serializable.Attribute,
        _expression.Any,
    ]

    for root in public_roots:
        expected_types = {root} | set(_collect_descendants(root))
        for t in expected_types:
            assert t.__name__ in dir(pydsdl), "Data type %r is not exported" % t



# Copyright (c) 2018 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import re
import string
from typing import Pattern, List, Union
from ._serializable import TypeParameterError


class InvalidNameError(TypeParameterError):
    pass


def check_name(name: str) -> None:
    """
    Ensure that the name complies with the requirements set out in the Specification;
    raise :class:`InvalidNameError` if not.
    """
    if not name:
        raise InvalidNameError("Name or namespace component cannot be empty")

    name = name.lower()

    if name[0] not in _VALID_FIRST_CHARACTERS_OF_NAME:
        raise InvalidNameError("Name or namespace component cannot start with %r" % name[0])

    for char in name:
        if char not in _VALID_CONTINUATION_CHARACTERS_OF_NAME:
            raise InvalidNameError("Name or namespace component cannot contain %r" % char)

    for pat in _DISALLOWED_NAME_PATTERNS:
        if isinstance(pat, str):
            if pat == name:
                raise InvalidNameError("Disallowed name: %r matches the following string: %s" % (name, pat))
        elif pat.match(name):
            raise InvalidNameError("Disallowed name: %r matches the following pattern: %s" % (name, pat))


_VALID_FIRST_CHARACTERS_OF_NAME = string.ascii_letters + "_"
_VALID_CONTINUATION_CHARACTERS_OF_NAME = _VALID_FIRST_CHARACTERS_OF_NAME + string.digits

# Disallowed name patterns apply to any part of any name, e.g., an attribute name, a namespace component,
# type name, etc. The pattern must produce an exact match to trigger a name error. All patterns are case-insensitive.
_DISALLOWED_NAME_PATTERNS: List[Union[str, Pattern[str]]] = [
    "truncated",
    "saturated",
    "true",
    "false",
    "bool",
    re.compile(r"void\d*$"),
    re.compile(r"u?int\d*$"),
    re.compile(r"u?q\d+_\d+$"),
    re.compile(r"float\d*$"),
    "optional",
    "aligned",
    "const",
    "struct",
    "super",
    "template",
    "enum",
    "self",
    "and",
    "or",
    "not",
    "auto",
    "type",
    "con",
    "prn",
    "aux",
    "nul",
    re.compile(r"com\d$"),
    re.compile(r"lpt\d$"),
    re.compile(r"_.*_$"),
]


def _unittest_check_name() -> None:
    from pytest import raises

    check_name("abc")
    check_name("_abc")
    check_name("abc_")
    check_name("abc0")

    with raises(InvalidNameError):
        check_name("0abc")

    with raises(InvalidNameError):
        check_name("_abc_")

    with raises(InvalidNameError):
        check_name("a-bc")

    with raises(InvalidNameError):
        check_name("")

    with raises(InvalidNameError):
        check_name("truncated")

    with raises(InvalidNameError):
        check_name("COM1")

    with raises(InvalidNameError):
        check_name("Aux")

    with raises(InvalidNameError):
        check_name("float128")

    with raises(InvalidNameError):
        check_name("q16_8")

    with raises(InvalidNameError):
        check_name("uq1_32")



# Copyright (c) 2018 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

# pylint: disable=consider-using-in,protected-access,too-many-statements

import fractions
from . import _any, _primitive, _container, _operator


# noinspection PyUnresolvedReferences,PyTypeChecker
def _unittest_expressions() -> None:
    r = _primitive.Rational
    s = _primitive.String

    for a in (True, False):
        for b in (True, False):
            assert _primitive.Boolean(a).native_value == a
            assert _operator.logical_not(_primitive.Boolean(a)).native_value == (not a)
            assert _operator.logical_and(_primitive.Boolean(a), _primitive.Boolean(b)).native_value == (a and b)
            assert _operator.logical_or(_primitive.Boolean(a), _primitive.Boolean(b)).native_value == (a or b)

    assert _operator.equal(
        _operator.divide(_operator.multiply(_operator.add(r(2), r(2)), r(3)), r(5)), r(fractions.Fraction(12, 5))
    ).native_value

    assert _operator.add(s("123"), s("abc")).native_value == "123abc"  # type: ignore

    new_set = _operator.add(_container.Set([s("123"), s("456")]), s("abc"))
    assert set(new_set) == {s("123abc"), s("456abc")}  # type: ignore

    new_set = _operator.add(s("abc"), _container.Set([s("123"), s("456")]))
    assert set(new_set) == {s("abc123"), s("abc456")}  # type: ignore

    new_set = _operator.add(
        s("abc"), _container.Set([_container.Set([s("123"), s("456")]), _container.Set([s("789"), s("987")])])
    )
    assert new_set == _container.Set(
        [_container.Set([s("abc123"), s("abc456")]), _container.Set([s("abc789"), s("abc987")])]
    )

    assert _operator.attribute(_container.Set([r(1), r(2), r(3), r(-4), r(-5)]), s("min")) == r(-5)
    assert _operator.attribute(_container.Set([r(1), r(2), r(3), r(-4), r(-5)]), s("max")) == r(3)


def _unittest_textual_representations() -> None:
    assert str(_primitive.Rational(fractions.Fraction(123, 456))) == "41/152"
    assert repr(_primitive.Rational(fractions.Fraction(123, 456))) == "rational(41/152)"
    assert str(_primitive.Rational(-123)) == "-123"
    assert repr(_primitive.Rational(-123)) == "rational(-123)"

    assert str(_primitive.Boolean(True)) == "true"
    assert repr(_primitive.Boolean(False)) == "bool(false)"

    assert str(_primitive.String("Hello\nworld!")) == r"'Hello\nworld!'"
    assert repr(_primitive.String("Hello\nworld!")) == r"string('Hello\nworld!')"

    tmp = str(_container.Set([_primitive.Rational(1), _primitive.Rational(fractions.Fraction(-9, 7))]))
    assert tmp == "{1, -9/7}" or tmp == "{-9/7, 1}"

    tmp = repr(_container.Set([_primitive.Rational(1), _primitive.Rational(fractions.Fraction(-9, 7))]))
    assert tmp == "set({1, -9/7})" or tmp == "set({-9/7, 1})"

    tmp = str(
        _container.Set(
            [
                _container.Set([_primitive.Rational(1), _primitive.Rational(fractions.Fraction(-9, 7))]),
                _container.Set([_primitive.Rational(fractions.Fraction(90, 7))]),
            ]
        )
    )
    assert (
        tmp == "{{1, -9/7}, {90/7}}"
        or tmp == "{{-9/7, 1}, {90/7}}"
        or tmp == "{{90/7}, {-9/7, 1}}"
        or tmp == "{{90/7}, {1, -9/7}}"
    )

    assert repr(_container.Set([_primitive.String("123")])) == "set({'123'})"


# noinspection PyTypeChecker
def _unittest_basic() -> None:
    from pytest import raises

    assert hash(_primitive.Boolean(True)) == hash(True)
    assert _primitive.Boolean(True) == _primitive.Boolean(True)
    assert _primitive.Boolean(True) != _primitive.Boolean(False)
    assert _primitive.Boolean(True) != _primitive.Rational(1)  # sic!
    assert _primitive.Boolean(True) != _primitive.Rational(123)
    assert _primitive.Boolean(True) != _container.Set([_primitive.Boolean(True)])

    with raises(ValueError):
        _primitive.Boolean(int)  # type: ignore

    with raises(ValueError):
        _primitive.Rational({123})  # type: ignore

    with raises(ValueError):
        _primitive.Rational("123")  # type: ignore

    with raises(ValueError):
        _primitive.String(123)  # type: ignore

    with raises(ValueError):
        _container.Set([123])  # type: ignore

    assert _primitive.Rational(123).is_integer()
    assert not _primitive.Rational(fractions.Fraction(123, 124)).is_integer()
    assert _primitive.Rational(-123).as_native_integer() == -123
    with raises(_any.InvalidOperandError):
        _primitive.Rational(fractions.Fraction(123, 124)).as_native_integer()



# Copyright (c) 2018 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

# pylint: disable=protected-access

import typing
import functools
from . import _any, _primitive


OperatorOutput = typing.TypeVar("OperatorOutput")
BinaryOperator = typing.Callable[[_any.Any, _any.Any], OperatorOutput]
AttributeOperator = typing.Callable[[_any.Any, typing.Union[_primitive.String, str]], OperatorOutput]


#
# Operator wrappers. These wrappers serve two purposes:
#   - Late binding, as explained here: https://stackoverflow.com/questions/55148139/referring-to-a-pure-virtual-method
#   - Automatic left-right operand swapping when necessary (for some polyadic operators).
#
def _auto_swap(
    alternative_operator_name: typing.Optional[str] = None,
) -> typing.Callable[[BinaryOperator[OperatorOutput]], BinaryOperator[OperatorOutput]]:
    def decorator(direct_operator: BinaryOperator[OperatorOutput]) -> BinaryOperator[OperatorOutput]:
        if alternative_operator_name:
            alternative_method_name = "_" + alternative_operator_name
        else:
            alternative_method_name = "_%s_right" % direct_operator.__name__

        if not hasattr(_any.Any, alternative_method_name):  # pragma: no cover
            raise TypeError("The following alternative operator method is not defined: %r" % alternative_method_name)

        @functools.wraps(direct_operator)
        def wrapper(left: _any.Any, right: _any.Any) -> OperatorOutput:
            if not isinstance(left, _any.Any) or not isinstance(right, _any.Any):  # pragma: no cover
                raise ValueError(
                    "Operators are only defined for implementations of Any; found this: %r, %r"
                    % (type(left).__name__, type(right).__name__)
                )
            try:
                result = direct_operator(left, right)
            except _any.UndefinedOperatorError:
                if type(left) != type(right):  # pylint: disable=unidiomatic-typecheck
                    result = getattr(right, alternative_method_name)(left)  # Left and Right are swapped.
                else:
                    raise

            assert isinstance(result, _any.Any)
            return typing.cast(OperatorOutput, result)

        return wrapper

    return decorator


def logical_not(operand: _any.Any) -> _primitive.Boolean:  # noinspection PyProtectedMember
    result = operand._logical_not()
    assert isinstance(result, _primitive.Boolean)
    return result


def positive(operand: _any.Any) -> _any.Any:  # noinspection PyProtectedMember
    return operand._positive()


def negative(operand: _any.Any) -> _any.Any:  # noinspection PyProtectedMember
    return operand._negative()


@_auto_swap("logical_or")  # Commutative
def logical_or(left: _any.Any, right: _any.Any) -> _primitive.Boolean:  # noinspection PyProtectedMember
    result = left._logical_or(right)
    assert isinstance(result, _primitive.Boolean)
    return result


@_auto_swap("logical_and")  # Commutative
def logical_and(left: _any.Any, right: _any.Any) -> _primitive.Boolean:  # noinspection PyProtectedMember
    result = left._logical_and(right)
    assert isinstance(result, _primitive.Boolean)
    return result


@_auto_swap("equal")  # Commutative
def equal(left: _any.Any, right: _any.Any) -> _primitive.Boolean:  # noinspection PyProtectedMember
    result = left._equal(right)
    assert isinstance(result, _primitive.Boolean)
    return result


# Special case - synthetic operator.
def not_equal(left: _any.Any, right: _any.Any) -> _primitive.Boolean:  # noinspection PyProtectedMember
    result = logical_not(equal(left, right))
    assert isinstance(result, _primitive.Boolean)
    return result


@_auto_swap("greater_or_equal")
def less_or_equal(left: _any.Any, right: _any.Any) -> _primitive.Boolean:  # noinspection PyProtectedMember
    result = left._less_or_equal(right)
    assert isinstance(result, _primitive.Boolean)
    return result


@_auto_swap("less_or_equal")
def greater_or_equal(left: _any.Any, right: _any.Any) -> _primitive.Boolean:  # noinspection PyProtectedMember
    result = left._greater_or_equal(right)
    assert isinstance(result, _primitive.Boolean)
    return result


@_auto_swap("greater")
def less(left: _any.Any, right: _any.Any) -> _primitive.Boolean:  # noinspection PyProtectedMember
    result = left._less(right)
    assert isinstance(result, _primitive.Boolean)
    return result


@_auto_swap("less")
def greater(left: _any.Any, right: _any.Any) -> _primitive.Boolean:  # noinspection PyProtectedMember
    result = left._greater(right)
    assert isinstance(result, _primitive.Boolean)
    return result


@_auto_swap()
def bitwise_or(left: _any.Any, right: _any.Any) -> _any.Any:  # noinspection PyProtectedMember
    return left._bitwise_or(right)


@_auto_swap()
def bitwise_xor(left: _any.Any, right: _any.Any) -> _any.Any:  # noinspection PyProtectedMember
    return left._bitwise_xor(right)


@_auto_swap()
def bitwise_and(left: _any.Any, right: _any.Any) -> _any.Any:  # noinspection PyProtectedMember
    return left._bitwise_and(right)


@_auto_swap()
def add(left: _any.Any, right: _any.Any) -> _any.Any:  # noinspection PyProtectedMember
    return left._add(right)


@_auto_swap()
def subtract(left: _any.Any, right: _any.Any) -> _any.Any:  # noinspection PyProtectedMember
    return left._subtract(right)


@_auto_swap()
def multiply(left: _any.Any, right: _any.Any) -> _any.Any:  # noinspection PyProtectedMember
    return left._multiply(right)


@_auto_swap()
def divide(left: _any.Any, right: _any.Any) -> _any.Any:  # noinspection PyProtectedMember
    return left._divide(right)


@_auto_swap()
def modulo(left: _any.Any, right: _any.Any) -> _any.Any:  # noinspection PyProtectedMember
    return left._modulo(right)


@_auto_swap()
def power(left: _any.Any, right: _any.Any) -> _any.Any:  # noinspection PyProtectedMember
    return left._power(right)


# Special case - no argument-swapped alternative defined.
# We accept both native strings and String in order to support both dynamically computed attributes and
# statically defined attributes.
def attribute(value: _any.Any, name: typing.Union[str, _primitive.String]) -> _any.Any:
    if isinstance(name, str):
        name = _primitive.String(name)

    if isinstance(value, _any.Any) and isinstance(name, _primitive.String):  # noinspection PyProtectedMember
        return value._attribute(name)

    raise ValueError(  # pragma: no cover
        "The argument types of the attribute operator are (Any, String), got (%r, %r)"
        % (type(value).__name__, type(name).__name__)
    )



# Copyright (c) 2018 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import typing
import warnings
from ._symbolic import Operator, NullaryOperator, MemoizationOperator


class BitLengthSet:
    """
    This type represents the Bit Length Set as defined in the Specification.
    It is used for representing bit offsets of fields and bit lengths of serialized representations.

    Most of the methods are evaluated analytically in nearly constant time rather than numerically.
    This is critical for complex layouts where numerical methods break due to combinatorial explosion and/or memory
    limits (see this discussed in https://github.com/UAVCAN/pydsdl/issues/23).
    There are several methods that trigger numerical expansion of the solution;
    due to the aforementioned combinatorial difficulties, they may be effectively incomputable in reasonable time,
    so production systems should not rely on them.

    Instances are guaranteed to be immutable.

    >>> b = 16 + BitLengthSet(8).repeat_range(256)
    >>> b
    BitLengthSet(concat({16},repeat(<=256,{8})))
    >>> b = 32 + b.repeat_range(65536)
    >>> b
    BitLengthSet(concat({32},repeat(<=65536,concat({16},repeat(<=256,{8})))))
    >>> b.min, b.max
    (32, 135266336)
    >>> sorted(b % 16)
    [0, 8]
    >>> sorted(b % 32)
    [0, 8, 16, 24]
    """

    def __init__(self, value: typing.Union[typing.Iterable[int], int, Operator, "BitLengthSet"]):
        """
        Accepts any iterable that yields integers (like another bit length set) or a single integer.
        """
        if isinstance(value, BitLengthSet):
            self._op = value._op  # type: Operator
        elif isinstance(value, Operator):
            self._op = MemoizationOperator(value)
        elif isinstance(value, int):
            self._op = NullaryOperator([value])
        else:
            self._op = NullaryOperator(value)

    # ========================================  QUERY METHODS  ========================================

    def is_aligned_at(self, bit_length: int) -> bool:
        """
        Shorthand for ``set(self % bit_length) == {0}``.

        >>> BitLengthSet(64).is_aligned_at(32)
        True
        >>> BitLengthSet(48).is_aligned_at(32)
        False
        >>> BitLengthSet(48).is_aligned_at(16)
        True
        >>> BitLengthSet(0).is_aligned_at(1234567)
        True
        """
        return set(self % bit_length) == {0}

    def is_aligned_at_byte(self) -> bool:
        """
        A shorthand for :meth:`is_aligned_at` using the standard byte size as prescribed by the Specification.

        >>> BitLengthSet(32).is_aligned_at_byte()
        True
        >>> BitLengthSet(33).is_aligned_at_byte()
        False
        """
        from .._serializable import SerializableType

        return self.is_aligned_at(SerializableType.BITS_PER_BYTE)

    @property
    def min(self) -> int:
        """
        The smallest element in the set derived analytically.

        >>> BitLengthSet.concatenate([{1, 2, 3}, {4, 5, 6}, {7, 8, 9}]).min
        12
        """
        return self._op.min

    @property
    def max(self) -> int:
        """
        The largest element in the set derived analytically.

        >>> BitLengthSet.concatenate([{1, 2, 3}, {4, 5, 6}, {7, 8, 9}]).pad_to_alignment(8).max
        24
        """
        return self._op.max

    @property
    def fixed_length(self) -> bool:
        """
        Shorthand for ``self.min == self.max``.

        >>> BitLengthSet(8).repeat(1).fixed_length
        True
        >>> BitLengthSet(8).repeat_range(1).fixed_length
        False
        """
        return self.min == self.max

    def __mod__(self, divisor: int) -> typing.Iterable[int]:
        """
        Elementwise modulus derived analytically.

        >>> sorted(BitLengthSet([0]) % 12345)
        [0]
        >>> sorted(BitLengthSet([8, 12, 16]) % 8)
        [0, 4]
        """
        # The type is reported as iterable[int], not sure yet if we should specialize it further. Time will tell.
        return BitLengthSet(self._op.modulo(int(divisor)))

    # ========================================  COMPOSITION METHODS  ========================================

    def pad_to_alignment(self, bit_length: int) -> "BitLengthSet":
        """
        Transform the bit length set expression such that the set becomes aligned at the specified alignment goal.
        After this transformation is applied, elements may become up to ``bit_length-1`` bits larger.
        The argument shall be a positive integer, otherwise it's a :class:`ValueError`.

        >>> from random import randint
        >>> alignment = randint(1, 64)
        >>> BitLengthSet(randint(1, 1000) for _ in range(100)).pad_to_alignment(alignment).is_aligned_at(alignment)
        True
        """
        from ._symbolic import PaddingOperator

        return BitLengthSet(PaddingOperator(self._op, bit_length))

    def repeat(self, k: int) -> "BitLengthSet":
        """
        Construct a new bit length set expression that repeats the current one the specified number of times.
        This reflects the arrangement of fixed-length DSDL array elements.
        This is a special case of :meth:`concatenate`.

        >>> sorted(BitLengthSet(1).repeat(0))
        [0]
        >>> sorted(BitLengthSet(1).repeat(1))
        [1]
        >>> sorted(BitLengthSet({1, 2, 3}).repeat(1))
        [1, 2, 3]
        >>> sorted(BitLengthSet({1, 2, 3}).repeat(2))
        [2, 3, 4, 5, 6]
        """
        from ._symbolic import RepetitionOperator

        return BitLengthSet(RepetitionOperator(self._op, k))

    def repeat_range(self, k_max: int) -> "BitLengthSet":
        """
        This is like :meth:`repeat` but ``k`` spans the range ``[0, k_max]``.

        >>> sorted(BitLengthSet({1, 2, 3}).repeat_range(2))
        [0, 1, 2, 3, 4, 5, 6]
        """
        from ._symbolic import RangeRepetitionOperator

        return BitLengthSet(RangeRepetitionOperator(self._op, k_max))

    @staticmethod
    def concatenate(sets: typing.Iterable[typing.Union["BitLengthSet", typing.Iterable[int], int]]) -> "BitLengthSet":
        """
        Construct a new bit length set expression that concatenates multiple bit length sets one after another.
        This reflects the data fields arrangement in a DSDL structure type.

        >>> sorted(BitLengthSet.concatenate([1, 2, 10]))
        [13]
        >>> sorted(BitLengthSet.concatenate([{1, 2}, {4, 5}]))
        [5, 6, 7]
        >>> sorted(BitLengthSet.concatenate([{1, 2, 3}, {4, 5, 6}]))
        [5, 6, 7, 8, 9]
        >>> sorted(BitLengthSet.concatenate([{1, 2, 3}, {4, 5, 6}, {7, 8, 9}]))
        [12, 13, 14, 15, 16, 17, 18]
        """
        from ._symbolic import ConcatenationOperator

        op = ConcatenationOperator(BitLengthSet(s)._op for s in sets)  # pylint: disable=protected-access
        return BitLengthSet(op)

    @staticmethod
    def unite(sets: typing.Iterable[typing.Union["BitLengthSet", typing.Iterable[int], int]]) -> "BitLengthSet":
        """
        Construct a new bit length set expression that is a union of multiple bit length sets.
        This reflects the data fields arrangement in a DSDL discriminated union.

        >>> sorted(BitLengthSet.unite([1, 2, 10]))
        [1, 2, 10]
        >>> sorted(BitLengthSet.unite([{1, 2}, {2, 3}]))
        [1, 2, 3]
        """
        from ._symbolic import UnionOperator

        op = UnionOperator(BitLengthSet(s)._op for s in sets)  # pylint: disable=protected-access
        return BitLengthSet(op)

    def __add__(self, other: typing.Union["BitLengthSet", typing.Iterable[int], int]) -> "BitLengthSet":
        """
        A shorthand for ``concatenate([self, other])``.
        One can easily see that if the argument is a set of one value (or a scalar),
        this method will result in the addition of said scalar to every element of the original set.

        >>> sorted(BitLengthSet(0) + BitLengthSet(0))
        [0]
        >>> sorted(BitLengthSet(4) + BitLengthSet(3))
        [7]
        >>> sorted(BitLengthSet({4, 91}) + 3)
        [7, 94]
        >>> sorted(BitLengthSet({4, 91}) + {5, 7})
        [9, 11, 96, 98]
        """
        return BitLengthSet.concatenate([self, other])

    def __radd__(self, other: typing.Union["BitLengthSet", typing.Iterable[int], int]) -> "BitLengthSet":
        """
        See :meth:`__add__`.

        >>> sorted({1, 2, 3} + BitLengthSet({4, 5, 6}))
        [5, 6, 7, 8, 9]
        >>> sorted(1 + BitLengthSet({2, 5, 7}))
        [3, 6, 8]
        """
        return BitLengthSet.concatenate([other, self])

    def __or__(self, other: typing.Union["BitLengthSet", typing.Iterable[int], int]) -> "BitLengthSet":
        """
        A shorthand for ``unite([self, other])``.

        >>> a = BitLengthSet(0)
        >>> a = a | BitLengthSet({1, 2, 3})
        >>> sorted(a)
        [0, 1, 2, 3]
        >>> a = a | {3, 4, 5}
        >>> sorted(a)
        [0, 1, 2, 3, 4, 5]
        >>> sorted(a | 6)
        [0, 1, 2, 3, 4, 5, 6]
        """
        return BitLengthSet.unite([self, other])

    def __ror__(self, other: typing.Union["BitLengthSet", typing.Iterable[int], int]) -> "BitLengthSet":
        """
        See :meth:`__or__`.

        >>> sorted({1, 2, 3} | BitLengthSet({4, 5, 6}))
        [1, 2, 3, 4, 5, 6]
        >>> sorted(1 | BitLengthSet({2, 5, 7}))
        [1, 2, 5, 7]
        """
        return BitLengthSet.unite([other, self])

    # ========================================  SLOW NUMERICAL METHODS  ========================================

    def __iter__(self) -> typing.Iterator[int]:
        """
        ..  attention::
            This method triggers slow numerical expansion.

            You might be tempted to use ``min(foo)`` or ``max(foo)`` for detecting length bounds.
            This may be effectively incomputable for data types with complex layout.
            Instead, use :attr:`min` and :attr:`max`.
        """
        return iter(self._op.expand())

    def __len__(self) -> int:
        """
        ..  attention::
            This method triggers slow numerical expansion.

            You might be tempted to use something like ``len(foo) == 1`` for detecting fixed-length sets.
            This may be effectively incomputable for data types with complex layout.
            Instead, use :attr:`fixed_length`.

        >>> len(BitLengthSet(0))
        1
        >>> len(BitLengthSet([1, 2, 3]))
        3
        """
        return len(self._op.expand())

    # ========================================  AUXILIARY METHODS  ========================================

    def __eq__(self, other: typing.Any) -> bool:
        """
        Currently, this method performs an approximate comparison that may yield a false-positive for some operands.
        This is done to avoid performing the costly numerical expansion of the operands.
        The implementation may be changed to perform exact comparison in the future if the underlying solver is
        updated accordingly.

        >>> BitLengthSet([1, 2, 4]) == {1, 2, 4}
        True
        >>> BitLengthSet([1, 2, 4]) == {1, 3, 4}
        False
        >>> BitLengthSet([123]) == BitLengthSet(123)
        True
        """
        try:
            other = BitLengthSet(other)
        except TypeError:
            return NotImplemented
        divisor = 32
        return self.min == other.min and self.max == other.max and set(self % divisor) == set(other % divisor)

    def __hash__(self) -> int:
        """
        Hash is computed in constant time (numerical expansion is not performed).

        >>> hash(BitLengthSet({1, 4})) != hash(BitLengthSet({1, 3}))
        True
        """
        return hash((self.min, self.max))

    def __bool__(self) -> bool:
        """
        This method is overridden to avoid accidental invocation of :meth:`__len__` in boolean expressions
        because it triggers numerical expansion.

        :return: Always True.
        """
        return True  # pragma: no cover

    def __str__(self) -> str:
        return str(self._op)

    def __repr__(self) -> str:
        return "%s(%s)" % (type(self).__name__, self)

    # ========================================  DEPRECATED METHODS   ========================================

    def elementwise_sum_k_multicombinations(self, k: int) -> "BitLengthSet":  # pragma: no cover
        """
        :meta private:
        """
        warnings.warn("Use repeat() instead", DeprecationWarning)
        return self.repeat(k)

    @staticmethod
    def elementwise_sum_cartesian_product(
        sets: typing.Iterable[typing.Union[typing.Iterable[int], int]]
    ) -> "BitLengthSet":  # pragma: no cover
        """
        :meta private:
        """
        warnings.warn("Use concatenate() instead", DeprecationWarning)
        return BitLengthSet.concatenate(sets)


def _unittest_bit_length_set() -> None:
    from pytest import raises

    assert BitLengthSet(0) == BitLengthSet(0)
    assert not (BitLengthSet(0) != BitLengthSet(0))  # pylint: disable=unneeded-not
    assert BitLengthSet(123) == BitLengthSet([123])
    assert BitLengthSet(123) != BitLengthSet(124)
    assert BitLengthSet(123) == 123
    assert BitLengthSet(123) != 124
    assert not (BitLengthSet(123) == "123")  # pylint: disable=unneeded-not
    assert str(BitLengthSet(0)) == "{0}"
    assert str(BitLengthSet(123)) == "{123}"
    assert str(BitLengthSet((123, 0, 456, 12))) == "{0,12,123,456}"  # Always sorted!
    assert BitLengthSet(0).is_aligned_at(1)
    assert BitLengthSet(0).is_aligned_at(1024)
    assert BitLengthSet(8).is_aligned_at_byte()
    assert not BitLengthSet(8).is_aligned_at(16)

    s = BitLengthSet({0, 8})
    s += 8
    assert s == {8, 16}
    s = s + {0, 4, 8}
    assert s == {8, 16, 12, 20, 24}

    assert BitLengthSet(0) + BitLengthSet(0) == BitLengthSet(0)
    assert BitLengthSet(4) + BitLengthSet(3) == {7}
    assert BitLengthSet({4, 91}) + 3 == {7, 94}
    assert BitLengthSet(7) + {12, 15} == {19, 22}
    assert {1, 2, 3} + BitLengthSet([4, 5, 6]) == {5, 6, 7, 8, 9}

    with raises(TypeError):
        assert BitLengthSet([4, 5, 6]) + "a"  # type: ignore

    with raises(TypeError):
        assert "a" + BitLengthSet([4, 5, 6])  # type: ignore

    with raises(TypeError):
        assert "a" | BitLengthSet([4, 5, 6])  # type: ignore

    with raises(TypeError):
        assert BitLengthSet([4, 5, 6]) | "a"  # type: ignore

    with raises(ValueError):
        BitLengthSet([4, 5, 6]).pad_to_alignment(0)



# Copyright (c) 2021 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import os
import abc
import math
import typing
import logging
import itertools


class Operator(abc.ABC):
    """
    Operators are immutable. This allows for aggressive caching and reference-sharing.
    """

    @abc.abstractmethod
    def modulo(self, divisor: int) -> typing.Set[int]:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def min(self) -> int:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def max(self) -> int:
        raise NotImplementedError

    @abc.abstractmethod
    def expand(self) -> typing.Set[int]:
        """
        Transform the symbolic form into numerical form.
        This is useful for cross-checking derived solutions and for DSDL expression evaluation.
        For complex expressions this may be incomputable due to combinatorial explosion or memory limits.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def __repr__(self) -> str:
        raise NotImplementedError


class NullaryOperator(Operator):
    """
    A nullary operator represents a constant value, which is a leaf of the operator tree.
    """

    def __init__(self, values: typing.Iterable[int]) -> None:
        self._value = set(values)
        if not self._value:
            raise ValueError("A bit length set cannot be empty. Did you mean to pass {0}?")
        for x in self._value:
            if not isinstance(x, int):
                raise TypeError("Invalid element for nullary set operator: %r" % x)

    def modulo(self, divisor: int) -> typing.Set[int]:
        return set(map(lambda x: x % divisor, self._value))

    @property
    def min(self) -> int:
        return min(self._value)

    @property
    def max(self) -> int:
        return max(self._value)

    def expand(self) -> typing.Set[int]:
        return set(self._value)

    def __repr__(self) -> str:
        return "{%s}" % ",".join(str(x) for x in sorted(self._value))


class PaddingOperator(Operator):
    """
    Adds up to ``alignment - 1`` padding bits to each entry of the child to ensure that the values are aligned.
    """

    def __init__(self, child: Operator, alignment: int) -> None:
        if alignment < 1:
            raise ValueError("Invalid alignment: %r bits" % alignment)
        self._child = child
        self._padding = int(alignment)

    def modulo(self, divisor: int) -> typing.Set[int]:
        r = self._padding
        mx = self.max
        lcm = least_common_multiple(r, divisor)
        out = set()  # type: typing.Set[int]
        for x in self._child.modulo(lcm):
            assert x <= mx and x < lcm
            out.add(self._pad(x) % divisor)
        return out

    @property
    def min(self) -> int:
        return self._pad(self._child.min)

    @property
    def max(self) -> int:
        return self._pad(self._child.max)

    def expand(self) -> typing.Set[int]:
        return set(map(self._pad, self._child.expand()))

    def _pad(self, x: int) -> int:
        r = self._padding
        return ((x + r - 1) // r) * r

    def __repr__(self) -> str:
        return "pad(%d,%r)" % (self._padding, self._child)


class ConcatenationOperator(Operator):
    """
    Given a set of children, transforms them into a single bit length set expression where each item is the
    elementwise sum of the cartesian product of the children's bit length sets.
    """

    def __init__(self, children: typing.Iterable[Operator]) -> None:
        self._children = list(children)
        if not self._children:
            raise ValueError("This operator is not defined on zero operands")

    def modulo(self, divisor: int) -> typing.Set[int]:
        # Take the modulus from each child and find all combinations.
        # The computational complexity is tightly bounded because the cardinality of the modulus set is less than
        # the bit length operand.
        mods = [ch.modulo(divisor) for ch in self._children]
        prod = itertools.product(*mods)
        sums = set(map(sum, prod))
        return {typing.cast(int, x) % divisor for x in sums}

    @property
    def min(self) -> int:
        return sum(x.min for x in self._children)

    @property
    def max(self) -> int:
        return sum(x.max for x in self._children)

    def expand(self) -> typing.Set[int]:
        return {sum(el) for el in itertools.product(*(x.expand() for x in self._children))}

    def __repr__(self) -> str:
        return "concat(%s)" % ",".join(map(repr, self._children))


class RepetitionOperator(Operator):
    """
    Concatenates ``k`` copies of the child.
    This is equivalent to :class:`ConcatenationOperator` where the child is replicated ``k`` times.
    """

    def __init__(self, child: Operator, k: int) -> None:
        self._k = int(k)
        self._child = child

    def modulo(self, divisor: int) -> typing.Set[int]:
        # Values of k > divisor will yield repeated entries so we can apply a reduction,
        # but the equivalent k modulo divisor shall be the same as k modulo divisor.
        # This is because the solution doesn't converge at k -> inf:
        #   {sum(x) % 2 for x in combinations_with_replacement({1, 3}, 2)} == {0}
        #   {sum(x) % 2 for x in combinations_with_replacement({1, 3}, 3)} == {1}
        #   {sum(x) % 2 for x in combinations_with_replacement({1, 3}, 4)} == {0}
        #   {sum(x) % 2 for x in combinations_with_replacement({1, 3}, 5)} == {1}
        # etc.
        equivalent_k = min(self._k, divisor + self._k % divisor)
        assert (self._k % divisor) == (equivalent_k % divisor), (divisor, self._k)
        return {
            (sum(el) % divisor)
            for el in itertools.combinations_with_replacement(self._child.modulo(divisor), equivalent_k)
        }

    @property
    def min(self) -> int:
        return self._child.min * self._k

    @property
    def max(self) -> int:
        return self._child.max * self._k

    def expand(self) -> typing.Set[int]:
        return {sum(el) for el in itertools.combinations_with_replacement(self._child.expand(), self._k)}

    def __repr__(self) -> str:
        return "repeat(%d,%r)" % (self._k, self._child)


class RangeRepetitionOperator(Operator):
    """
    Concatenates ``k in [0, k_max]`` copies of the child.
    In other words, this is like ``k+1`` instances of :class:`RepetitionOperator`.
    """

    def __init__(self, child: Operator, k_max: int) -> None:
        self._k_max = int(k_max)
        self._child = child

    def modulo(self, divisor: int) -> typing.Set[int]:
        single = self._child.modulo(divisor)
        assert isinstance(single, set)
        # Values of k > divisor will yield repeated entries so we can apply a reduction (see non-range case above).
        # This holds only if the argument does not contain repeated entries which is guaranteed by `set`.
        equivalent_k_max = min(self._k_max, divisor + self._k_max % divisor)
        assert (self._k_max % divisor) == (equivalent_k_max % divisor), (divisor, self._k_max)
        out = set()  # type: typing.Set[int]
        for k in range(equivalent_k_max + 1):
            for el in itertools.combinations_with_replacement(single, k):
                out.add(sum(el) % divisor)
        return out

    @property
    def min(self) -> int:
        return 0

    @property
    def max(self) -> int:
        return self._child.max * self._k_max

    def expand(self) -> typing.Set[int]:
        ch = self._child.expand()
        assert isinstance(ch, set)
        out = set()  # type: typing.Set[int]
        for k in range(self._k_max + 1):
            for el in itertools.combinations_with_replacement(ch, k):
                out.add(sum(el))
        return out

    def __repr__(self) -> str:
        return "repeat(<=%d,%r)" % (self._k_max, self._child)


class UnionOperator(Operator):
    def __init__(self, children: typing.Iterable[Operator]) -> None:
        self._children = list(children)
        if not self._children:
            raise ValueError("This operator is not defined on zero operands")

    def modulo(self, divisor: int) -> typing.Set[int]:
        out = set()  # type: typing.Set[int]
        for x in self._children:
            out |= x.modulo(divisor)
        return out

    @property
    def min(self) -> int:
        return min(x.min for x in self._children)

    @property
    def max(self) -> int:
        return max(x.max for x in self._children)

    def expand(self) -> typing.Set[int]:
        out = set()  # type: typing.Set[int]
        for x in self._children:
            out |= x.expand()
        return out

    def __repr__(self) -> str:
        return "(%s)" % "|".join(map(repr, self._children))


class MemoizationOperator(Operator):
    """
    This is a no-op transparent lazy cache on top of the child operator.
    It also logs a stack trace if the child takes too long to expand to help with bottleneck optimization.
    """

    def __init__(self, child: Operator) -> None:
        self._child = child
        self._min = None  # type: typing.Optional[int]
        self._max = None  # type: typing.Optional[int]
        self._modula = {}  # type: typing.Dict[int, typing.Set[int]]
        self._expansion = None  # type: typing.Optional[typing.Set[int]]

    def modulo(self, divisor: int) -> typing.Set[int]:
        try:
            return self._modula[divisor]
        except LookupError:
            self._modula[divisor] = self._child.modulo(divisor)
        return self._modula[divisor]

    @property
    def min(self) -> int:
        if self._min is None:
            self._min = self._child.min
        return self._min

    @property
    def max(self) -> int:
        if self._max is None:
            self._max = self._child.max
        return self._max

    def expand(self) -> typing.Set[int]:
        if self._expansion is None:
            from time import monotonic

            # Track the time and log occurrences that take a long time to help find bottlenecks in user code
            # that accidentally relies on numerical expansion. This is mainly intended to help us transition
            # Nunavut to the new solver API instead of numerical methods. It may be removed later.
            started_at = monotonic()
            self._expansion = self._child.expand()
            elapsed = monotonic() - started_at
            if elapsed > 2.0:  # pragma: no cover
                _logger.info(
                    "Numerical expansion took %.1f seconds; the result contains %d items:\n%s",
                    elapsed,
                    len(self._expansion),
                    self._child,
                    stack_info=True,
                )
            assert elapsed < _POISON_SLOW_EXPANSION_SECONDS

            # Since we did an expansion anyway, the set must be compact,
            # so we use this opportunity to validate the correctness of the solver.
            # This may be removed easily since it has no visible effects.
            validate_numerically(self)

        return self._expansion

    def __repr__(self) -> str:
        return repr(self._child)  # Not sure if we should indicate our presence considering that we're a no-op


def least_common_multiple(a: int, b: int) -> int:
    """
    This replicates :func:`math.lcm` to support Python <3.9.
    """
    return abs(a * b) // math.gcd(a, b)


def validate_numerically(op: Operator) -> None:
    """
    Validates the correctness of symbolic derivations by comparing the results against reference values
    obtained numerically.
    The computational complexity may be prohibitively high for some inputs due to combinatorial explosion.
    In case of a divergence the function triggers an assertion failure.
    """
    s = op.expand()
    assert min(s) == op.min
    assert max(s) == op.max
    for div in range(1, 65):
        assert op.modulo(div) == {x % div for x in s}, div


_POISON_SLOW_EXPANSION_SECONDS = float(os.environ.get("PYDSDL_POISON_SLOW_EXPANSION_SECONDS", "999999999"))
"""
This is intended for developers only so it is not mentioned in the public documentation.
The purpose is to trigger an assertion failure if a numerical expansion takes more than this many seconds.
"""

_logger = logging.getLogger(__name__)



# Copyright (c) 2018 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

# pylint: disable=broad-except

import typing
from pathlib import Path
import urllib.parse


class FrontendError(Exception):  # PEP8 says that the "Exception" suffix is redundant and should not be used.
    """
    This is the root exception type for all custom exceptions defined in the library.
    This type itself is not expected to be particularly useful to the library user;
    please refer to the direct descendants instead.
    """

    def __init__(self, text: str, path: typing.Optional[Path] = None, line: typing.Optional[int] = None):
        Exception.__init__(self, text)
        self._path = path
        self._line = line

    def set_error_location_if_unknown(
        self, path: typing.Optional[Path] = None, line: typing.Optional[int] = None
    ) -> None:
        """
        Entries that are already known will be left unchanged.
        This is useful when propagating exceptions through recursive instances,
        e.g., when processing nested definitions.
        """
        if not self._path and path:
            self._path = path

        if not self._line and line:
            self._line = line

    @property
    def path(self) -> typing.Optional[Path]:
        """Source file path where the error has occurred, if known."""
        return self._path

    @property
    def line(self) -> typing.Optional[int]:
        """
        Source file line number (first line numbered 1) where the error has occurred, if known.
        The path is always known if the line number is set.
        """
        return self._line

    @property
    def text(self) -> str:
        return Exception.__str__(self)

    def __str__(self) -> str:
        """
        Nicely formats an error string in the typical error format ``[file:[line:]]description``.
        Example::

            uavcan/internet/udp/500.HandleIncomingPacket.1.0.uavcan:33: Error such and such
        """
        if self.path and self.line:
            return "%s:%d: %s" % (self.path.as_posix(), self.line, self.text)

        if self.path:
            return "%s: %s" % (self.path.as_posix(), self.text)

        return self.text

    def __repr__(self) -> str:
        return self.__class__.__name__ + ": " + repr(self.__str__())


class InternalError(FrontendError):
    """
    This exception is used to report internal errors in the front end itself that prevented it from
    processing the definitions. Every occurrence should be reported to the developers.
    """

    def __init__(
        self,
        text: typing.Optional[str] = None,
        path: typing.Optional[Path] = None,
        line: typing.Optional[int] = None,
        culprit: typing.Optional[Exception] = None,
    ):
        if culprit is not None:
            report_text = "PLEASE REPORT AT https://github.com/UAVCAN/pydsdl/issues/new?title=" + urllib.parse.quote(
                repr(culprit)
            )
            if text:
                text = text + " " + report_text
            else:  # pragma: no cover
                text = report_text

        if not text:
            text = ""

        super().__init__(text=text, path=path, line=line)


class InvalidDefinitionError(FrontendError):
    """
    This exception type is used to point out mistakes and errors in DSDL definitions.
    This type is inherited by a dozen of specialized exception types; however, the class hierarchy beneath
    this type may be unstable and should not be relied upon by the application directly.
    """


def _unittest_error() -> None:
    try:
        raise FrontendError("Hello world!")
    except Exception as ex:
        assert str(ex) == "Hello world!"
        assert repr(ex) == "FrontendError: 'Hello world!'"

    try:
        raise FrontendError("Hello world!", path=Path("path/to/file.uavcan"), line=123)
    except Exception as ex:
        assert str(ex) == "path/to/file.uavcan:123: Hello world!"
        assert repr(ex) == "FrontendError: 'path/to/file.uavcan:123: Hello world!'"

    try:
        raise FrontendError("Hello world!", path=Path("path/to/file.uavcan"))
    except Exception as ex:
        assert str(ex) == "path/to/file.uavcan: Hello world!"
        assert repr(ex) == "FrontendError: 'path/to/file.uavcan: Hello world!'"


def _unittest_internal_error_github_reporting() -> None:
    try:
        raise InternalError(path=Path("FILE_PATH"), line=42)
    except FrontendError as ex:
        assert ex.path == Path("FILE_PATH")
        assert ex.line == 42
        assert str(ex) == "FILE_PATH:42: "

    try:
        try:
            try:  # TRY HARDER
                raise InternalError(text="BASE TEXT", culprit=Exception("ERROR TEXT"))
            except FrontendError as ex:
                ex.set_error_location_if_unknown(path=Path("FILE_PATH"))
                raise
        except FrontendError as ex:
            ex.set_error_location_if_unknown(line=42)
            raise
    except FrontendError as ex:
        print(ex)
        assert ex.path == Path("FILE_PATH")
        assert ex.line == 42
        # We have to ignore the last couple of characters because Python before 3.7 reprs Exceptions like this:
        #   Exception('ERROR TEXT',)
        # But newer Pythons do it like this:
        #   Exception('ERROR TEXT')
        assert str(ex).startswith(
            "FILE_PATH:42: BASE TEXT "
            "PLEASE REPORT AT https://github.com/UAVCAN/pydsdl/issues/new?title=Exception%28%27ERROR%20TEXT%27"
        )

    try:
        raise InternalError(text="BASE TEXT", path=Path("FILE_PATH"))
    except FrontendError as ex:
        assert str(ex) == "FILE_PATH: BASE TEXT"



#!/usr/bin/env python3
# Copyright (c) 2020 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>
# type: ignore

import sys
import setuptools

if int(setuptools.__version__.split(".")[0]) < 30:
    print(
        "A newer version of setuptools is required. The current version does not support declarative config.",
        file=sys.stderr,
    )
    sys.exit(1)

setuptools.setup()



# Copyright (c) 2018 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import abc
import math
import typing
from .._bit_length_set import BitLengthSet
from ._serializable import SerializableType, TypeParameterError
from ._primitive import UnsignedIntegerType, PrimitiveType


class InvalidNumberOfElementsError(TypeParameterError):
    pass


class ArrayType(SerializableType):
    def __init__(self, element_type: SerializableType, capacity: int):
        super().__init__()
        self._element_type = element_type
        self._capacity = int(capacity)
        if self._capacity < 1:
            raise InvalidNumberOfElementsError("Array capacity cannot be less than 1")

    @property
    def element_type(self) -> SerializableType:
        return self._element_type

    @property
    def capacity(self) -> int:
        """
        The (maximum) number of elements in the (variable-length) array.
        """
        return self._capacity

    @property
    def string_like(self) -> bool:
        """
        True if the array might contain a text string, in which case it is termed to be "string-like".
        A string-like array is a variable-length array of ``uint8``.
        See https://github.com/UAVCAN/specification/issues/51.
        """
        return False

    @property
    def alignment_requirement(self) -> int:
        """
        The alignment requirement of an array equals that of its element type.
        The length of the serialized representation of any type is a multiple of its alignment requirement;
        therefore, every element is always placed such that its alignment requirement is satisfied.
        """
        return self.element_type.alignment_requirement

    @abc.abstractmethod
    def __str__(self) -> str:  # pragma: no cover
        raise NotImplementedError


class FixedLengthArrayType(ArrayType):
    def __init__(self, element_type: SerializableType, capacity: int):
        super().__init__(element_type, capacity)
        self._bls = self.element_type.bit_length_set.repeat(self.capacity)
        assert self._bls.is_aligned_at(self.alignment_requirement)

    @property
    def bit_length_set(self) -> BitLengthSet:
        return self._bls

    def enumerate_elements_with_offsets(
        self, base_offset: BitLengthSet = BitLengthSet(0)
    ) -> typing.Iterator[typing.Tuple[int, BitLengthSet]]:
        """
        This is a convenience method for code generation.
        Its behavior mimics that of :meth:`pydsdl.StructureType.iterate_fields_with_offsets`,
        except that we iterate over indexes instead of fields.

        :param base_offset: The base offset to add to each element. If not supplied, assumed to be ``{0}``.
            The base offset will be implicitly padded out to :attr:`alignment_requirement`.

        :returns: For an N-element array, an iterator over N elements, where each element is a tuple of the index
            of the array element (zero-based) and its offset as a bit length set.
        """
        base_offset = base_offset.pad_to_alignment(self.alignment_requirement)
        for index in range(self.capacity):
            offset = base_offset + self.element_type.bit_length_set.repeat(index)
            assert offset.is_aligned_at(self.element_type.alignment_requirement)
            yield index, offset

    def __str__(self) -> str:
        return "%s[%d]" % (self.element_type, self.capacity)

    def __repr__(self) -> str:
        return "FixedLengthArrayType(element_type=%r, capacity=%r)" % (self.element_type, self.capacity)


def _unittest_fixed_array() -> None:
    from pytest import raises
    from ._primitive import SignedIntegerType

    su8 = UnsignedIntegerType(8, cast_mode=PrimitiveType.CastMode.TRUNCATED)
    ti64 = SignedIntegerType(64, cast_mode=PrimitiveType.CastMode.SATURATED)

    assert str(FixedLengthArrayType(su8, 4)) == "truncated uint8[4]"
    assert str(FixedLengthArrayType(ti64, 1)) == "saturated int64[1]"

    assert not FixedLengthArrayType(su8, 4).string_like
    assert not FixedLengthArrayType(ti64, 1).string_like

    assert FixedLengthArrayType(su8, 4).bit_length_set == 32
    assert FixedLengthArrayType(su8, 200).capacity == 200
    assert FixedLengthArrayType(ti64, 200).element_type is ti64

    with raises(InvalidNumberOfElementsError):
        FixedLengthArrayType(ti64, 0)

    assert (
        repr(FixedLengthArrayType(ti64, 128))
        == "FixedLengthArrayType(element_type=SignedIntegerType(bit_length=64, cast_mode=<CastMode.SATURATED: 0>), "
        "capacity=128)"
    )

    small = FixedLengthArrayType(su8, 2)
    assert small.bit_length_set == {16}
    assert list(small.enumerate_elements_with_offsets()) == [(0, BitLengthSet(0)), (1, BitLengthSet(8))]


class VariableLengthArrayType(ArrayType):
    def __init__(self, element_type: SerializableType, capacity: int):
        super().__init__(element_type, capacity)

        # Construct the implicit array length prefix type.
        length_field_length = 2 ** math.ceil(math.log2(max(self.BITS_PER_BYTE, self.capacity.bit_length())))

        # If the length field is less than the alignment requirement (which, at the time of writing this,
        # is not possible because the max alignment is 8 and the min length length is also 8),
        # it would break the alignment of the array elements. Hence, we ensure that it is never smaller.
        length_field_length = max(length_field_length, self.alignment_requirement)
        assert length_field_length % self.element_type.alignment_requirement == 0

        self._length_field_type = UnsignedIntegerType(length_field_length, PrimitiveType.CastMode.TRUNCATED)

        self._bls = self.length_field_type.bit_length + self.element_type.bit_length_set.repeat_range(self.capacity)
        assert self._bls.is_aligned_at(self.alignment_requirement)

    @property
    def bit_length_set(self) -> BitLengthSet:
        assert self._bls.is_aligned_at(self.alignment_requirement)
        return self._bls

    @property
    def string_like(self) -> bool:
        """See the base class."""
        et = self.element_type  # Without this temporary MyPy yields a false positive type error
        return isinstance(et, UnsignedIntegerType) and (et.bit_length == self.BITS_PER_BYTE)

    @property
    def length_field_type(self) -> UnsignedIntegerType:
        """
        The unsigned integer type of the implicit array length field.
        Note that the set of valid length values is a subset of that of the returned type.
        """
        assert self._length_field_type.bit_length % self.element_type.alignment_requirement == 0
        return self._length_field_type

    def __str__(self) -> str:
        return "%s[<=%d]" % (self.element_type, self.capacity)

    def __repr__(self) -> str:
        return "VariableLengthArrayType(element_type=%r, capacity=%r)" % (self.element_type, self.capacity)


def _unittest_variable_array() -> None:
    from pytest import raises
    from ._primitive import SignedIntegerType

    tu8 = UnsignedIntegerType(8, cast_mode=PrimitiveType.CastMode.TRUNCATED)
    si64 = SignedIntegerType(64, cast_mode=PrimitiveType.CastMode.SATURATED)

    assert str(VariableLengthArrayType(tu8, 4)) == "truncated uint8[<=4]"
    assert str(VariableLengthArrayType(si64, 255)) == "saturated int64[<=255]"

    assert VariableLengthArrayType(tu8, 4).string_like
    assert not VariableLengthArrayType(si64, 1).string_like

    # Mind the length prefix!
    assert VariableLengthArrayType(tu8, 3).bit_length_set == {8, 16, 24, 32}
    assert VariableLengthArrayType(tu8, 1).bit_length_set == {8, 16}
    assert max(VariableLengthArrayType(tu8, 255).bit_length_set) == 2048

    assert VariableLengthArrayType(tu8, 200).capacity == 200
    assert VariableLengthArrayType(tu8, 200).element_type is tu8

    with raises(InvalidNumberOfElementsError):
        VariableLengthArrayType(si64, 0)

    assert (
        repr(VariableLengthArrayType(si64, 128))
        == "VariableLengthArrayType(element_type=SignedIntegerType(bit_length=64, cast_mode=<CastMode.SATURATED: 0>), "
        "capacity=128)"
    )

    small = VariableLengthArrayType(tu8, 2)
    assert small.bit_length_set == {8, 16, 24}

    outer = FixedLengthArrayType(small, 2)
    assert outer.bit_length_set == {16, 24, 32, 40, 48}

    assert VariableLengthArrayType(tu8, 100).length_field_type.bit_length == 8
    assert VariableLengthArrayType(tu8, 10000).length_field_type.bit_length == 16
    assert VariableLengthArrayType(tu8, 1000000).length_field_type.bit_length == 32
    assert VariableLengthArrayType(tu8, 10000000000).length_field_type.bit_length == 64
